#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
PyDP4 integrated workflow for the running of MM, DFT GIAO calculations and
DP4 analysis
v1.0

Copyright (c) 2015-2019 Kristaps Ermanis, Alexander Howarth, Jonathan M. Goodman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

Created on Wed Nov 19 15:26:32 2014
Updated on Feb 07 2019

@author: ke291

The main file, that should be called to start the PyDP4 workflow.
Interprets the arguments and takes care of the general workflow logic.
"""

import sys
import os
import datetime
import argparse
import importlib
import getpass
from pathlib import Path
import warnings
warnings.filterwarnings("ignore")

import NMR
import DP5 as DP5
import DP4 as DP4
import ConfPrune

MMpackages = [['m', 't', 'e'],['Macromodel','Tinker', 'ETKDG']]
DFTpackages = [['n', 'w', 'g', 'z', 'd'],['NWChem', 'NWChemZiggy', 'Gaussian', 'GaussianZiggy', 'GaussianDarwin']]


# Assigning the config default values
# Settings are defined roughly in the order they are used in the script
class Settings:
    # --- Main options ---
    MM = 'm'  # m for MacroModel, t for Tinker
    DFT = 'z'  # n, g, z or for NWChem or Gaussian
    Workflow = 'gmns'  # defines which steps to include in the workflow
    # c for RDkit cleaning of input structures and 3d coordinate generation
    # g for generate diastereomers
    # m for molecular mechanics conformational search
    # o for DFT optimization
    # e for DFT single-point energies
    # n for DFT NMR calculation
    # s for computational and experimental NMR data extraction and stats analysis
    # w for DP5 probability calculation
    Solvent = ''  # solvent for DFT optimization and NMR calculation
    ScriptDir = ''  # Script directory, automatically set on launch
    InputFiles = []  # Structure input files - can be MacroModel *-out.mae or *sdf files
    InputFilesPaths = []  # Path object for Structure input files - can be MacroModel *-out.mae or *sdf files
    NMRsource = ''  # File or folder containing NMR description or data
    Title = 'DP4molecule'  # Title of the calculation, set to NMR file name by default on launch
    AssumeDone = False  # Assume all computations done, only read DFT output data and analyze (use for reruns)
    AssumeConverged = False  # Assume all optimizations have converged, do NMR and/or energy calcs on existing DFT geometries
    UseExistingInputs = False  # Don't regenerate DFT inputs, use existing ones. Good for restarting a failed calc
    Smiles = None  # Smiles input file - text file with smiles strings on separate lines
    InChIs = None  # InChI input file - text file with inchi strings on separate lines
    Smarts = None  # Smarts input file - text file with Smarts strings on separate lines

    # --- Diastereomer generation ---
    SelectedStereocentres = []  # which stereocentres to vary for diastereomer generation

    # --- Molecular mechanics ---
    ForceField = 'mmff'  # ff tfOPto use for conformational search
    MMstepcount = 10000  # Max number of MM steps to do, if less than MMfactor*rotable_bonds
    MMfactor = 2500  # MMfactor*rotable_bonds gives number of steps to do if less than MMstepcount
    Rot5Cycle = False  # Special dealing with 5-membered saturated rings, see FiveConf.py
    RingAtoms = []  # Define the 5-membered ring, useful if several are present in molecule
    SCHRODINGER = ''  # Define the root folder for Schrodinger software
    TinkerPath = '/home/ah809/tinker'  # Define the root folder for Tinker software,
    # must contain bin/scan and params/mmff.prm for the process to work

    # --- Conformer pruning ---
    HardConfLimit = 1000  # Immediately stop if conformers to run exceed this number
    ConfPrune = True  # Should we prune conformations?
    PerStructConfLimit = 100  # Max numbers of conformers allowed per structure for DFT stages
    InitialRMSDcutoff = 0.75  # Initial RMSD threshold for pruning
    MaxCutoffEnergy = 10.0  # Max conformer MM energy in kJ/mol to allow

    # --- DFT ---
    NWChemPath = "nwchem"  # Path to nwchem executable. If it's in the path, can be just 'nwchem'
    GausPath = "g09"  # Path to Gaussian executable. If it's in the path, can be just 'g09' or 'g16'
    # If left empty, it will attempt to use g09 in GAUS_EXEDIR environment variable
    MaxDFTOptCycles = 50  # Max number of DFT geometry optimization cycles to request.
    CalcFC = False  # Calculate QM force constants before optimization
    OptStepSize = 30  # Max step Gaussian should take in geometry optimization
    charge = None  # Manually specify charge for DFT calcs
    nBasisSet = "6-311g(d)"  # Basis set for NMR calcs
    nFunctional = "mPW1PW91"  # Functional for NMR calcs
    oBasisSet = "6-31g(d,p)"  # Basis set for geometry optimizations
    oFunctional = "b3lyp"  # Functional for geometry optimizations
    eBasisSet = "def2tzvp"  # Basis set for energy calculations
    eFunctional = "m062x"  # Functional for energy calculations

    # --- Computational clusters ---
    """ These should probably be moved to relevant *.py files as Cambridge specific """
    user = ''  # Linux user on computational clusters, not used for local calcs
    TimeLimit = 24  # Queue time limit on comp clusters
    queue = 'SWAN'  # Which queue to use on Ziggy
    project = 'GOODMAN-SL3-CPU'  # Which project to use on Darwin
    DarwinScrDir = '/home/u/rds/hpc-work/'  # Which scratch directory to use on Darwin
    StartTime = ''  # Automatically set on launch, used for folder names
    nProc = 1  # Cores used per job, must be less than node size on cluster
    DarwinNodeSize = 32  # Node size on current CSD3
    MaxConcurrentJobsZiggy = 75  # Max concurrent jobs to submit on ziggy
    MaxConcurrentJobsDarwin = 320  # Max concurrent jobs to submit on CSD3

    # --- NMR analysis ---
    TMS_SC_C13 = 191.69255  # Default TMS reference C shielding constant (from B3LYP/6-31g**)
    TMS_SC_H1 = 31.7518583  # Default TMS reference H shielding constant (from B3LYP/6-31g**)

    # --- Stats ---
    StatsModel = 'g'  # What statistical model type to use
    StatsParamFile = 'none'  # Where to find statistical model parameters

    # --- Output folder ---
    OutputFolder = ''  # folder to print dp4 output to - default is cwd
    GUIRunning = False  # Boolean has PyDP4 been called from commandline or from GUI


settings = Settings()


# Data structure keeping all of isomer data in one place.
class Isomer:
    def __init__(self, InputFile, Charge=-100):
        self.InputFile = InputFile  # Initial structure input file
        self.BaseName = InputFile  # Basename for other files
        self.Atoms = []  # Element labels
        self.Conformers = []  # from conformational search, list of atom coordinate lists
        self.MMCharge = 0  # charge from conformational search
        self.ExtCharge = Charge  # externally provided charge
        self.RMSDCutoff = 0  # RMSD cutoff eventually used to get the conformer number below the limit
        self.DFTConformers = []  # from DFT optimizations, list of atom coordinate lists
        self.ConfIndices = []  # List of conformer indices from the original conformational search for reference
        self.MMEnergies = []  # Corresponding MM energies in kj/mol
        self.DFTEnergies = []  # Corresponding DFT energies in hartrees
        self.Energies = []  # Final energies used in conformer population prediction in kj/mol
        self.Populations = []  # Conformer populations
        self.OptInputFiles = []  # list of DFT NMR input file names
        self.OptOutputFiles = []  # list of DFT NMR output file names
        self.EInputFiles = []  # list of DFT NMR input file names
        self.EOutputFiles = []  # list of DFT NMR output file names
        self.NMRInputFiles = []  # list of DFT NMR input file names
        self.NMROutputFiles = []  # list of DFT NMR output file names
        self.ShieldingLabels = []  # A list of atom labels corresponding to the shielding values
        self.ConformerShieldings = []  # list of calculated NMR shielding constant lists for every conformer
        self.ConformerCShifts = [] # list of calculated C NMR shifts lists for every conformer
        self.ConformerHShifts = [] # list of calculated H NMR shifts lists for every conformer
        self.BoltzmannShieldings = []  # Boltzmann weighted NMR shielding constant list for the isomer
        self.Cshifts = []  # Calculated C NMR shifts
        self.Hshifts = []  # Calculated H NMR
        self.Clabels = []
        self.Hlabels = []
        self.Mols = []
        self.Cexp = []  # Experimental C NMR shifts
        self.Hexp = []  # Experimental H NMR shifts

def main(settings):

    print("Current working directory: " + os.getcwd())
    print("Initial input files: " + str(settings.InputFiles))
    print("NMR file: " + str(settings.NMRsource))
    print("Workflow: " + str(settings.Workflow))

    # Read in any text inputs and add these to the input file list

    import StructureInput

    if settings.Smiles:
        settings.InputFiles.extend(StructureInput.GenerateSDFFromTxt(settings.Smiles, 'Smiles'))

    if settings.Smarts:
        settings.InputFiles.extend(StructureInput.GenerateSDFFromTxt(settings.Smarts, 'Smarts'))

    if settings.InChIs:
        settings.InputFiles.extend(StructureInput.GenerateSDFFromTxt(settings.InChIs, 'InChI'))

    # Clean up input files if c in workflow - this generates a new set of 3d coordinates as a starting point

    if 'c' in settings.Workflow or ("m" not in settings.Workflow and "o" not in settings.Workflow)  and len(settings.InputFiles) > 0:
        import StructureInput

        # if requested generate 3d coordinates to define any stereochemistry

        settings.InputFiles = StructureInput.CleanUp(settings.InputFiles)

    # if no structure inputs have been found at this point quit

    if len(settings.InputFiles) == 0:
        print("\nNo input files were found please use -h for help with input options quitting...")

        quit()

    # if g in workflow check number of stereocentres for each input and generate and diastereomers

    if ('g' in settings.Workflow):

        from DiastereomerGeneration import GenDiastereomers
        print("\nGenerating diastereomers...")

        FinalInputFiles = []

        if len(settings.InputFiles) > 1:
            settings.SelectedStereocentres = []
        
        for InpFile in settings.InputFiles:
            DiastereomerFilenames = GenDiastereomers(InpFile, protected_atoms=settings.SelectedStereocentres,double_bonds=True)
            FinalInputFiles.extend(DiastereomerFilenames)

        settings.InputFiles = list(FinalInputFiles)

    settings.InputFilesPaths = [Path.cwd() / i for i in settings.InputFiles]

    print("Generated input files: " + str(settings.InputFiles) + '\n')

    # Create isomer data structures
    Isomers = [Isomer(f.split('.sdf')[0]) for f in settings.InputFiles]

    print("Assuming all computations are done? ... ", settings.AssumeDone)
    print("Using preexisting DFT data? ... ", settings.UseExistingInputs)

    # Run conformational search, if requested
    if ('m' in settings.Workflow) and not (settings.AssumeDone or settings.UseExistingInputs):

        #print("Performing conformational search using ", end="")

        MM = ImportMM(settings.MM)
        print("Performing conformational search")
        print('\nSetting up files...')
        MMInputs = MM.SetupMM(settings)
        print('MM inputs: '+str(MMInputs))
        print('Running MM:')
        MMOutputs = MM.RunMM(MMInputs, settings)
        print('\nReading conformers...')
        Isomers = MM.ReadConformers(MMOutputs, Isomers, settings)
        for iso in Isomers:
            print(iso.InputFile + ": " + str(len(iso.Conformers)) + ' conformers read within energy window')
    else:
        print('No conformational search was requested. Skipping...')
        settings.ConfPrune = False

    # Prune conformations, if requested.
    # For each isomer, the conformers list is replaced with a smaller list of conformers
    if (settings.ConfPrune) and not (settings.AssumeDone or settings.UseExistingInputs):
        print('\nPruning conformers...')
        Isomers = ConfPrune.RMSDPrune(Isomers, settings)
        for iso in Isomers:
            print(iso.InputFile + ": " + str(len(iso.Conformers)) + ' conformers after pruning with ' +
                  str(iso.RMSDCutoff) + 'A RMSD cutoff')

    if ('n' in settings.Workflow) or ('o' in settings.Workflow) \
            or ('e' in settings.Workflow):
        DFT = ImportDFT(settings.DFT)
    else:
        print('\nNo DFT calculations were requested. Skipping...')

    if not (settings.AssumeDone):

        # Run DFT optimizations, if requested
        if ('o' in settings.Workflow):

            now = datetime.datetime.now()
            settings.StartTime = now.strftime('%d%b%H%M')

            print('\nSetting up geometry optimization calculations...')
            Isomers = DFT.SetupOptCalcs(Isomers, settings)
            print('\nRunning geometry optimization calculations...')
            Isomers = DFT.RunOptCalcs(Isomers, settings)

            print('\nReading DFT optimized geometries...')

            Isomers = DFT.ReadGeometries(Isomers, settings)

            # Add convergence check here before continuing with calcs!
            if (DFT.Converged(Isomers) == False) and (settings.AssumeConverged == False):
                print('Some of the conformers did not converge, quitting...')
                quit()

        # Run DFT single-point energy calculations, if requested
        if ('e' in settings.Workflow):

            now = datetime.datetime.now()
            settings.StartTime = now.strftime('%d%b%H%M')

            print('\nSetting up energy calculations...')
            Isomers = DFT.SetupECalcs(Isomers, settings)
            print('\nRunning energy calculations...')
            Isomers = DFT.RunECalcs(Isomers, settings)
            print('\nReading data from the output files...')
            Isomers = DFT.ReadEnergies(Isomers, settings)
            print("Energies: ")
            for iso in Isomers:
                print(iso.InputFile + ": " + str(iso.DFTEnergies))

        # Run DFT NMR calculations, if requested
        if ('n' in settings.Workflow):

            now = datetime.datetime.now()
            settings.StartTime = now.strftime('%d%b%H%M')

            print('\nSetting up NMR calculations...')
            Isomers = DFT.SetupNMRCalcs(Isomers, settings)
            print('\nRunning NMR calculations...')
            Isomers = DFT.RunNMRCalcs(Isomers, settings)
            print('\nReading data from the output files...')
            Isomers = DFT.ReadShieldings(Isomers)
            print("Shieldings: ")
            for iso in Isomers:
                print(iso.InputFile + ": ")
                for conf in iso.ConformerShieldings:
                    print(str(conf))

            Isomers = DFT.ReadEnergies(Isomers, settings)
            print("Energies: ")
            for iso in Isomers:
                print(iso.InputFile + ": " + str(iso.DFTEnergies))

    else:
        # Read DFT optimized geometries, if requested
        if ('o' in settings.Workflow):
            Isomers = DFT.GetPrerunOptCalcs(Isomers)
            Isomers = DFT.ReadGeometries(Isomers, settings)

        if ('e' in settings.Workflow):
            Isomers = DFT.GetPrerunECalcs(Isomers)
            Isomers = DFT.ReadEnergies(Isomers, settings)

        if ('n' in settings.Workflow):
            Isomers = DFT.GetPrerunNMRCalcs(Isomers)
            Isomers = DFT.ReadShieldings(Isomers)

    if ('n' in settings.Workflow):

        print('\nSetting TMS computational NMR shielding constant references')
        settings.TMS_SC_C13, settings.TMS_SC_H1 = NMR.GetTMSConstants(settings)

        print('\nConverting DFT data to NMR shifts...')
        Isomers = NMR.CalcBoltzmannWeightedShieldings(Isomers)
        Isomers = NMR.CalcNMRShifts(Isomers, settings)

    elif ('s' in settings.Workflow) or ('w' in settings.Workflow):

        #predict NMR shifts with an ML model instead

        Cshifts , Clabels, Isomers =  DP5.predict_Exp_shifts(settings,Isomers)

        for iso , cshifts,clabels in zip(Isomers,Cshifts,Clabels):

            iso.Cshifts = cshifts
            iso.Clabels = clabels

    if ('s' in settings.Workflow) or ('w' in settings.Workflow) or ("a" in settings.Workflow):

        print('\nReading experimental NMR data...')
        NMRData = NMR.NMRData(settings)

        """
        print("Conformation data:")
        NMR.PrintConformationData(Isomers)
        """

        if NMRData.Type == 'desc':

            print('Experimental NMR description found and read.')

            # performs a pairwise assignment
            Isomers = NMR.PairwiseAssignment(Isomers, NMRData)

            print('Cshifts: ' + str(NMRData.Cshifts))
            print('Hshifts: ' + str(NMRData.Hshifts))

            print('Equivalents: ' + str(NMRData.Equivalents))
            print('Omits: ' + str(NMRData.Omits))

        elif NMRData.Type == "fid":

            for f in settings.NMRsource:

                if f.name == "Proton" or f.name == "proton":

                    from Proton_assignment import AssignProton
                    from Proton_plotting import PlotProton

                    print('\nAssigning proton spectrum...')
                    Isomers = AssignProton(NMRData, Isomers, settings)

                    if settings.GUIRunning == False:
                        print('\nPlotting proton spectrum...')
                        PlotProton(NMRData, Isomers, settings)

                elif f.name == "Carbon" or f.name == "carbon":

                    from Carbon_assignment import AssignCarbon
                    from Carbon_plotting import PlotCarbon

                    print('\nAssigning carbon spectrum...')
                    Isomers = AssignCarbon(NMRData, Isomers, settings)

                    if settings.GUIRunning == False:
                        print('\nPlotting carbon spectrum...')
                        PlotCarbon(NMRData, Isomers, settings)

        elif NMRData.Type == "jcamp":

            for f in settings.NMRsource:

                if f.name == "Proton.dx" or f.name == "proton.dx":

                    from Proton_assignment import AssignProton
                    from Proton_plotting import PlotProton

                    print('\nAssigning proton spectrum...')
                    Isomers = AssignProton(NMRData, Isomers, settings)

                    if settings.GUIRunning == False:
                        print('\nPlotting proton spectrum...')
                        PlotProton(NMRData, Isomers, settings)

                elif f.name == "Carbon.dx" or f.name == "carbon.dx":

                    from Carbon_assignment import AssignCarbon
                    from Carbon_plotting import PlotCarbon

                    print('\nAssigning carbon spectrum...')
                    Isomers = AssignCarbon(NMRData, Isomers, settings)

                    if settings.GUIRunning == False:
                        print('\nPlotting carbon spectrum...')
                        PlotCarbon(NMRData, Isomers, settings)

            print('Raw FID NMR datafound and read.')

    # print('\nProcessing experimental NMR data...')

    # NMRdata = NMR.ProcessNMRData(Isomers, settings.NMRsource, settings)

    if 'w' in settings.Workflow:

        print('\nCalculating DP5 probabilities...')

        # make folder for WF data to go into

        DP5data = DP5.DP5data(Path(settings.ScriptDir), len(Isomers[0].Atoms))

        if not os.path.exists(Path(settings.OutputFolder)):
            os.mkdir(Path(settings.OutputFolder))

        if not os.path.exists(Path(settings.OutputFolder) / 'dp5'):
            os.mkdir(Path(settings.OutputFolder) / 'dp5')

            DP5data = DP5.ProcessIsomers(DP5data, Isomers, settings)

            if "n" in settings.Workflow:

                DP5data.ErrorAtomProbs = DP5.kde_probs(Isomers, settings, "Error", DP5data.ErrorAtomReps,DP5data.ConfCshifts, DP5data.Cexp)

                DP5data.B_ErrorAtomProbs = DP5.BoltzmannWeight_DP5(Isomers, DP5data.ErrorAtomProbs)

                DP5data.Mol_Error_probs = DP5.Calculate_DP5(DP5data.B_ErrorAtomProbs)

                DP5data.DP5_Error_probs, DP5data.B_ErrorAtomProbs = DP5.Rescale_DP5(DP5data.Mol_Error_probs,
                                                                                     DP5data.B_ErrorAtomProbs, settings,
                                                                                     "Error",DP5data.CMAE)

                final_ps = DP5data.DP5_Error_probs

            else:

                DP5data.ExpAtomProbs = DP5.kde_probs(Isomers, settings, "Exp", DP5data.ExpAtomReps, [], DP5data.Cexp)

                DP5data.B_ExpAtomProbs = DP5.BoltzmannWeight_DP5(Isomers, DP5data.ExpAtomProbs)

                DP5data.Mol_Exp_probs = DP5.Calculate_DP5(DP5data.B_ExpAtomProbs)

                DP5data.DP5_Exp_probs, DP5data.DP5ExpAtomProbs = DP5.Rescale_DP5(DP5data.Mol_Exp_probs,
                                                                                 DP5data.B_ExpAtomProbs, settings,
                                                                                 "Exp",DP5data.CMAE)

                final_ps = DP5data.DP5_Exp_probs

            print('Saving DP5 probabilities...')

            DP5data = DP5.Pickle_res(DP5data, settings)

        else:

            DP5data = DP5.UnPickle_res(DP5data, settings)

        if "n" in settings.Workflow:

            DP5data.Output = DP5.MakeOutput( Isomers, settings,DP5data,DP5data.DP5_Error_probs,DP5data.B_ErrorAtomProbs)

            f = open("/".join(str(settings.OutputFolder).split("/")[:-1]) + "/abs_mean_max_bayes", "a+")

            f.write( str(settings.NMRsource[0].name)+ " "+str(DP5data.CMAE) + " "  +str(DP5data.CMax)  + "\n")

            quit()

        else:

            DP5data.Output = DP5.MakeOutput( Isomers, settings,DP5data,DP5data.DP5_Exp_probs,DP5data.B_ExpAtomProbs)


    else:

        DP5data = []

    if ('s' in settings.Workflow):

        if len(Isomers) < 2:

            print("Multiple structures required for DP4 probability calculations...")

        else:

            print('\nCalculating DP4 probabilities...')

            DP4data = DP4.DP4data()
            DP4data = DP4.ProcessIsomers(DP4data, Isomers)
            DP4data = DP4.InternalScaling(DP4data)
            DP4data = DP4.CalcProbs(DP4data, settings)
            DP4data = DP4.CalcDP4(DP4data)

            DP4data = DP4.MakeOutput(DP4data, Isomers, settings)

            import pickle


    else:
        print('\nNo DP4 analysis requested.')

        DP4data = []

    print('\nPyDP4 process completed successfully.')

    print("workflow" , settings.Workflow)

    return NMRData, Isomers, settings, DP4data, DP5data


# Selects which MM package to import, returns imported module
def ImportMM(mm):
    if mm in MMpackages[0]:
        MMindex = MMpackages[0].index(mm)
        MM = importlib.import_module(MMpackages[1][MMindex])
    else:
        print("Invalid MM package selected")
        quit()
    return MM


# Selects which DFT package to import, returns imported module
def ImportDFT(dft):
    if dft in DFTpackages[0]:
        DFTindex = DFTpackages[0].index(dft)
        DFT = importlib.import_module(DFTpackages[1][DFTindex])
    else:
        print("Invalid DFT package selected")
        quit()

    return DFT


def getScriptPath():
    return os.path.dirname(os.path.realpath(sys.argv[0]))


def NMR_files(NMR_args):

    print("NMR_path")

    NMR_path = Path(NMR_args)
    NMR_Data = []

    # check if path is from cwd or elsewhere:
    if len(NMR_path.parts) == 1:
        # if so a folder in the cwd has been passed add the cwd to the path
        NMR_path = Path.cwd() / NMR_path
        print(NMR_path)

    # now check if it is a directory or a file, add proton and carbon data here
    if NMR_path.is_dir():

        p_switch = 0
        c_switch = 0

        for f in NMR_path.iterdir():

            if f.name == "Carbon" or f.name == "carbon" or f.name == "Carbon.dx" or f.name == "carbon.dx":
                NMR_Data.append(f)
                c_switch = 1

            elif f.name == "Proton" or f.name == "proton" or f.name == "Proton.dx" or f.name == "proton.dx":
                NMR_Data.append(f)
                p_switch = 1

            if p_switch == 1 and c_switch == 1:
                break

        # self.NMR_list.addItem(str(filename[0].split("/")[-1]))

        if p_switch == 0 and c_switch == 0:
            NMR_Data.append(f)

    # if its not a directory add the file

    else:

        NMR_Data.append(NMR_path)

    settings.NMRsource = NMR_Data

    return


# Read the config file and fill in the corresponding attributes in settings class
def ReadConfig(settings):
    cfgpath = os.path.join(getScriptPath(), 'settings.cfg')
    if not os.path.exists(cfgpath):
        print('settings.cfg is missing!')
        return settings

    configfile = open(cfgpath, 'r')
    config = configfile.readlines()
    configfile.close()

    # Read in the new settings values from config
    newsettings = []
    for line in config:
        if ('#' in line) or (len(line) < 3) or ('=' not in line):
            continue

        newsettings.append([x.strip() for x in line[:-1].split('=')])
        if len(newsettings[-1]) < 2:
            newsettings[-1].append('')

    # Set the attributes in the settings class
    print('Settings read from settings.cfg:')
    for setting in newsettings:
        if hasattr(settings, setting[0]):
            setattr(settings, setting[0], setting[1])
            print('  ' + setting[0] + ': ' + setting[1])

    return settings


if __name__ == '__main__':

    print("==========================")
    print("PyDP4 script,\nintegrating Tinker/MacroModel,")
    print("Gaussian/NWChem and DP4\nv1.0")
    print("\nCopyright (c) 2015-2019 Kristaps Ermanis, Alexander Howarth, Jonathan M. Goodman")
    print("Distributed under MIT license")
    print("==========================\n\n")

    # Read config file and fill in settings in from that
    settings = ReadConfig(settings)

    # These are then overridden by any explicit parameters given through the command line
    parser = argparse.ArgumentParser(description='PyDP4 script to setup\
    and run Tinker, Gaussian (on ziggy) and DP4')
    parser.add_argument('-w', '--workflow', help="Defines which steps to include in the workflow, " +
                                                 "can contain g for generate diastereomers, m for molecular mechanics conformational search, " +
                                                 "o for DFT optimization, e for DFT single-point energies, n for DFT NMR calculation, " +
                                                 "a for computational and experimental NMR data extraction " +
                                                 "s for computational and experimental NMR data extraction and stats analysis, default is 'gmns'",
                        default=settings.Workflow)
    parser.add_argument('-m', '--mm', help="Select molecular mechanics program,\
    t for tinker or m for macromodel, default is m", choices=['t', 'm', 'e'],
                        default='m')
    parser.add_argument('-d', '--dft', help="Select DFT program, \
    g for Gaussian, n for NWChem, z for Gaussian on ziggy, d for Gaussian on \
    Darwin, default is g", choices=DFTpackages[0], default='g')
    parser.add_argument('--StepCount', help="Specify\
    stereocentres for diastereomer generation")

    parser.add_argument('StructureFiles', nargs='*', default=[], help=
    "One or more SDF file for the structures to be verified by DP4. At least one\
    is required, if automatic diastereomer generation is used.")

    parser.add_argument("ExpNMR", help="Experimental NMR description, assigned\
    with the atom numbers from the structure file")

    parser.add_argument("-s", "--solvent", help="Specify solvent to use\
    for dft calculations")
    parser.add_argument("-q", "--queue", help="Specify queue for job submission\
    on ziggy", default=settings.queue)
    parser.add_argument("--TimeLimit", help="Specify job time limit for jobs\
    on ziggy or darwin", type=int)

    parser.add_argument("--nProc", help="Specify number of processor cores\
    to use for Gaussian calculations", type=int, default=1)
    parser.add_argument("--batch", help="Specify max number of jobs per batch",
                        type=int, default=settings.MaxConcurrentJobsZiggy)
    parser.add_argument("--project", help="Specify project for job submission\
    on darwin", default=settings.project)
    parser.add_argument("--ConfLimit", help="Specify maximum number of \
    conformers per structure. If above this, adaptive RMSD pruning will be \
    performed", type=int, default=settings.PerStructConfLimit)

    parser.add_argument("--MaxConfE", help="Specify maximum MMFF energy \
    allowed before conformer is discarded before DFT stage", type=float, \
                        default=settings.MaxCutoffEnergy)

    parser.add_argument("-r", "--rot5", help="Manually generate conformers for\
    5-memebered rings", action="store_true")

    parser.add_argument('--ra', help="Specify ring atoms, for the ring to be\
    rotated, useful for molecules with several 5-membered rings")
    parser.add_argument('-S', '--Stats', help="Specify the stats model and\
    parameters")

    parser.add_argument("--AssumeDFTDone", help="Assume RMSD pruning, DFT setup\
    and DFT calculations have been run already", action="store_true")
    parser.add_argument("--AssumeConverged", help="Assume DFT optimizations have" + \
                                                  " converged and can be used for NMR and or energy calcs",
                        action="store_true")
    parser.add_argument("--UseExistingInputs", help="Use previously generated\
    DFT inputs, avoids long conf pruning and regeneration", action="store_true")
    parser.add_argument("--NoConfPrune", help="Skip RMSD pruning, use all\
    conformers in the energy window", action="store_true")

    parser.add_argument('-c', '--StereoCentres', help="Specify\
    stereocentres for diastereomer generation")
    parser.add_argument("--OptCycles", help="Specify max number of DFT geometry\
    optimization cycles", type=int, default=settings.MaxDFTOptCycles)
    parser.add_argument("--OptStep", help="Specify the max step size\
    Gaussian should take in optimization, default is 30", type=int, default=settings.OptStepSize)
    parser.add_argument("--FC", help="Calculate force constants before optimization", action="store_true")

    parser.add_argument('-n', '--Charge', help="Specify\
    charge of the molecule. Do not use when input files have different charges")
    parser.add_argument('-B', '--nBasisSet', help="Selects the basis set for\
    DFT NMR calculations", default=settings.nBasisSet)
    parser.add_argument('-F', '--nFunctional', help="Selects the functional for\
    DFT NMR calculations", default=settings.nFunctional)
    parser.add_argument('--eBasisSet', help="Selects the basis set for\
    DFT energy calculations", default=settings.eBasisSet)
    parser.add_argument('--eFunctional', help="Selects the functional for\
    DFT energy calculations", default=settings.eFunctional)
    parser.add_argument('-f', '--ff', help="Selects force field for the \
    conformational search, implemented options 'mmff' and 'opls' (2005\
    version)", choices=['mmff', 'opls'], default=settings.ForceField)

    parser.add_argument('--OutputFolder', help="Directory for dp4 output default is cwd", default=settings.OutputFolder)

    parser.add_argument('--Smiles', help="txt file input containing smiles strings on separate lines",
                        default=settings.Smiles)

    parser.add_argument('--Smarts', help="txt file input containing smarts strings on separate lines",
                        default=settings.Smarts)

    parser.add_argument('--InChIs', help="txt file input containing inchi strings on separate lines",
                        default=settings.InChIs)

    args = parser.parse_args()
    print(args.StructureFiles)
    print(args.ExpNMR)

    settings.Title = args.ExpNMR
    settings.NMRsource = args.ExpNMR
    settings.Workflow = args.workflow

    settings.DFT = args.dft
    settings.queue = args.queue
    settings.ScriptDir = getScriptPath()
    settings.ForceField = args.ff
    settings.PerStructConfLimit = args.ConfLimit
    settings.MaxCutoffEnergy = args.MaxConfE
    settings.nBasisSet = args.nBasisSet
    settings.nFunctional = args.nFunctional
    settings.eBasisSet = args.eBasisSet
    settings.eFunctional = args.eFunctional
    settings.nProc = args.nProc
    settings.MaxConcurrentJobs = args.batch
    settings.project = args.project
    settings.MaxDFTOptCycles = args.OptCycles
    settings.OptStepSize = args.OptStep
    if args.FC:
        settings.CalcFC = True

    if args.TimeLimit:
        settings.TimeLimit = args.TimeLimit

    if args.Stats is not None:
        settings.StatsModel = (args.Stats)[0]
        settings.StatsParamFile = (args.Stats)[1:]

    settings.MM = args.mm

    if args.StepCount is not None:
        settings.MMstepcount = int(args.StepCount)
    if args.Charge is not None:
        settings.charge = int(args.Charge)
    if args.StereoCentres is not None:
        settings.SelectedStereocentres = \
            [int(x) for x in (args.StereoCentres).split(',')]
    if args.NoConfPrune:
        settings.ConfPrune = False
    if args.AssumeDFTDone:
        settings.AssumeDone = True
    if args.AssumeConverged:
        settings.AssumeConverged = True
    if args.UseExistingInputs:
        settings.UseExistingInputs = True
    if args.solvent:
        settings.Solvent = args.solvent
    if args.rot5:
        settings.Rot5Cycle = True
    if args.ra is not None:
        settings.RingAtoms = \
            [int(x) for x in (args.ra).split(',')]

    if settings.StatsParamFile != 'none':
        if os.path.isfile(settings.StatsParamFile):
            print("Statistical parameter file found at " + settings.StatsParamFile)
        elif (not os.path.isfile(settings.StatsParamFile)) and \
                os.path.isfile(settings.ScriptDir + settings.StatsParamFile):
            settings.StatsParamFile = settings.ScriptDir + settings.StatsParamFile
            print("Statistical parameter file found at " + settings.StatsParamFile)
        elif (not os.path.isfile(settings.StatsParamFile)) and \
                (not os.path.isfile(settings.ScriptDir + settings.StatsParamFile)):
            print("Stats file not found, quitting.")

    now = datetime.datetime.now()
    settings.StartTime = now.strftime('%d%b%H%M')

    settings.user = getpass.getuser()
    settings.DarwinScrDir.replace('/u/', settings.user)

    with open('cmd.log', 'a') as f:
        f.write(' '.join(sys.argv) + '\n')

    settings.InputFiles = args.StructureFiles

    settings.Smiles = args.Smiles
    settings.Smarts = args.Smarts
    settings.InChIs = args.InChIs

    settings.NMRsource = args.ExpNMR

    NMR_files(settings.NMRsource)

    # check if NMR data has been passed from the cwd or the full path

    settings.OutputFolder = Path(args.OutputFolder)

    main(settings)



"""
These classes wrap the GRUCell and LSTMCell keras layers, allowing weight-tying
between RNNs for each atom. Essentially, Keras expects inputs to an RNN to be
of the shape (batch_size, sequence_length, features), while our inputs are of
shape (num_atoms_per_batch, features). 

Peter, 3/20/2018

I should revisit whether these are needed post graph_nets update.

Peter, 11/9/2018.

"""

from keras.layers import GRUCell, LSTMCell

class GRUStep(GRUCell): 
    def build(self, input_shape):
        GRUCell.build(self, input_shape[0])

    def call(self, inputs):
        """ Inputs should be [message, previous_state], returns [next_state]
        """
        return GRUCell.call(self, inputs[0], [inputs[1]])[0]

    def compute_output_shape(self, input_shape):
        return input_shape[0]


class LSTMStep(LSTMCell): 
    def build(self, input_shape):
        LSTMCell.build(self, input_shape[0])

    def call(self, inputs):
        """ Inputs should be [message, previous_state, previous_memory], 
        returns [next_state, next_memory]
        """
        outputs = LSTMCell.call(self, inputs[0], [inputs[1], inputs[2]])
        return [outputs[0], outputs[1][1]]

    def compute_mask(self, inputs, mask):
        return [None] * 2
        
    def compute_output_shape(self, input_shape):
        return [input_shape[0], input_shape[0]]



#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
Created on Wed Nov 19 15:56:54 2014
Rewritten during April 2019

@author: ke291

Contains all of the Gaussian specific code for input generation and calculation
execution. Called by PyDP4.py.
"""

import subprocess
import os
import time
import glob
import shutil
import math
from subprocess import STDOUT, check_output
gasConstant = 8.3145
temperature = 298.15
hartreeEnergy = 2625.499629554010


def SetupNMRCalcs(Isomers, settings):

    jobdir = os.getcwd()

    if not os.path.exists('nmr'):
        os.mkdir('nmr')
    os.chdir('nmr')

    for iso in Isomers:
        if iso.ExtCharge > -10:
            charge = iso.ExtCharge
        else:
            charge = iso.MMCharge

        if iso.DFTConformers == []:
            conformers = iso.Conformers
        else:
            conformers = iso.DFTConformers

        for num in range(0, len(conformers)):
            filename = iso.BaseName + 'ginp' + str(num + 1).zfill(3)

            if os.path.exists(filename + '.out'):
                if IsGausCompleted(filename + '.out'):
                    iso.NMROutputFiles.append(filename + '.out')
                    continue
                else:
                    os.remove(filename + '.out')

            WriteGausFile(filename, conformers[num], iso.Atoms, charge, settings, 'nmr')
            iso.NMRInputFiles.append(filename + '.com')

    os.chdir(jobdir)

    return Isomers


def SetupECalcs(Isomers, settings):

    jobdir = os.getcwd()

    if not os.path.exists('e'):
        os.mkdir('e')
    os.chdir('e')

    for iso in Isomers:
        if iso.ExtCharge > -10:
            charge = iso.ExtCharge
        else:
            charge = iso.MMCharge

        if iso.DFTConformers == []:
            conformers = iso.Conformers
        else:
            conformers = iso.DFTConformers

        for num in range(0, len(conformers)):
            filename = iso.BaseName + 'ginp' + str(num + 1).zfill(3)

            if os.path.exists(filename + '.out'):
                if IsGausCompleted(filename + '.out'):
                    iso.EOutputFiles.append(filename + '.out')
                    continue
                else:
                    os.remove(filename + '.out')

            WriteGausFile(filename, conformers[num], iso.Atoms, charge, settings, 'e')
            iso.EInputFiles.append(filename + '.com')

    os.chdir(jobdir)

    return Isomers


def SetupOptCalcs(Isomers, settings):

    jobdir = os.getcwd()

    if not os.path.exists('opt'):
        os.mkdir('opt')
    os.chdir('opt')

    for iso in Isomers:
        if iso.ExtCharge > -10:
            charge = iso.ExtCharge
        else:
            charge = iso.MMCharge

        for num in range(0, len(iso.Conformers)):
            filename = iso.BaseName + 'ginp' + str(num + 1).zfill(3)

            if os.path.exists(filename + '.out'):
                if IsGausCompleted(filename + '.out'):
                    if IsGausConverged(filename + '.out') or (settings.AssumeConverged == True):
                        iso.OptOutputFiles.append(filename + '.out')
                        continue
                    else:
                        # If calculation completed, but didn't converge, reuse geometry and resubmit
                        atoms, coords = ReadGeometry(filename + '.out')
                        if coords != []:
                            print('Partially optimised structure found for ' + filename + ', reusing')
                            iso.Conformers[num] = coords
                        os.remove(filename + '.out')
                else:
                    os.remove(filename + '.out')

            WriteGausFile(filename, iso.Conformers[num], iso.Atoms, charge, settings, 'opt')
            iso.OptInputFiles.append(filename + '.com')

    os.chdir(jobdir)

    return Isomers


def Converged(Isomers):

    jobdir = os.getcwd()

    if not os.path.exists('opt'):
        os.chdir(jobdir)
        return False

    os.chdir('opt')

    for iso in Isomers:
        for num in range(0, len(iso.Conformers)):
            filename = iso.BaseName + 'ginp' + str(num + 1).zfill(3)

            if os.path.exists(filename + '.out'):
                if IsGausConverged(filename + '.out') == False:
                    os.chdir(jobdir)
                    return False
            else:
                os.chdir(jobdir)
                return False

    os.chdir(jobdir)
    return True


def RunNMRCalcs(Isomers, settings):

    print('\nRunning Gaussian DFT NMR calculations locally...')

    jobdir = os.getcwd()
    os.chdir('nmr')

    GausJobs = []

    for iso in Isomers:
        GausJobs.extend([x for x in iso.NMRInputFiles if (x[:-4] + '.out') not in iso.NMROutputFiles])

    Completed = RunCalcs(GausJobs, settings)

    for iso in Isomers:
        iso.NMROutputFiles.extend([x[:-4] + '.out' for x in iso.NMRInputFiles if (x[:-4] + '.out') in Completed])

    os.chdir(jobdir)

    return Isomers


def GetPrerunNMRCalcs(Isomers):

    print('\nLooking for prerun Gaussian DFT NMR files...')

    jobdir = os.getcwd()
    os.chdir('nmr')

    for iso in Isomers:
        iso.NMRInputFiles = glob.glob(iso.BaseName + 'ginp*com')
        iso.NMROutputFiles.extend([x[:-4] + '.out' for x in iso.NMRInputFiles if IsGausCompleted(x[:-4] + '.out')])

    print('NMR calc files:')
    print(', '.join([', '.join(x.NMROutputFiles) for x in Isomers]))

    os.chdir(jobdir)

    return Isomers


def RunECalcs(Isomers, settings):

    print('\nRunning Gaussian DFT energy calculations locally...')

    jobdir = os.getcwd()
    os.chdir('e')

    GausJobs = []

    for iso in Isomers:
        GausJobs.extend([x for x in iso.EInputFiles if (x[:-4] + '.out') not in iso.EOutputFiles])

    Completed = RunCalcs(GausJobs, settings)

    for iso in Isomers:
        iso.EOutputFiles.extend([x[:-4] + '.out' for x in iso.EInputFiles if (x[:-4] + '.out') in Completed])

    os.chdir(jobdir)

    return Isomers


def GetPrerunECalcs(Isomers):

    print('\nLooking for prerun Gaussian DFT energy calculation files...')

    jobdir = os.getcwd()
    os.chdir('e')

    for iso in Isomers:
        iso.EInputFiles = glob.glob(iso.BaseName + 'ginp*com')
        iso.EOutputFiles.extend([x[:-4] + '.out' for x in iso.EInputFiles if IsGausCompleted(x[:-4] + '.out')])

    print('Energy files:')
    print(', '.join([', '.join(x.EOutputFiles) for x in Isomers]))

    os.chdir(jobdir)

    return Isomers


def RunOptCalcs(Isomers, settings):

    print('\nRunning Gaussian DFT geometry optimizations locally...')

    jobdir = os.getcwd()
    os.chdir('opt')

    GausJobs = []

    for iso in Isomers:
        GausJobs.extend([x for x in iso.OptInputFiles if (x[:-4] + '.out') not in iso.OptOutputFiles])

    Completed = RunCalcs(GausJobs, settings)

    for iso in Isomers:
        iso.OptOutputFiles.extend([x[:-4] + '.out' for x in iso.OptInputFiles if (x[:-4] + '.out') in Completed])

    os.chdir(jobdir)

    return Isomers


def GetPrerunOptCalcs(Isomers):

    print('\nLooking for prerun Gaussian DFT optimization files...')

    jobdir = os.getcwd()
    os.chdir('opt')

    for iso in Isomers:
        iso.OptInputFiles = glob.glob(iso.BaseName + 'ginp*com')
        iso.OptOutputFiles.extend([x[:-4] + '.out' for x in iso.OptInputFiles if IsGausCompleted(x[:-4] + '.out')])

    print('Opt files:')
    print(', '.join([', '.join(x.OptOutputFiles) for x in Isomers]))

    os.chdir(jobdir)

    return Isomers


def RunCalcs(GausJobs, settings):

    NCompleted = 0
    Completed = []

    if len(GausJobs) == 0:
        print("There were no jobs to run.")
        return Completed

    if ('GAUS_EXEDIR' in os.environ):
        gausdir = os.environ['GAUSS_EXEDIR']
        if shutil.which(os.path.join(gausdir, 'g09')) is None:
            GausPrefix = os.path.join(gausdir, "g16")
        else:
            GausPrefix = os.path.join(gausdir, "g09")
    else:
        GausPrefix = settings.GausPath

    if shutil.which(GausPrefix) is None:
        print('Gaussian.py, RunCalcs:\n  Could not find Gaussian executable at ' + GausPrefix)
        quit()

    for f in GausJobs:
        time.sleep(3)
        print(GausPrefix + " < " + f + ' > ' + f[:-3] + 'out')
        outp = subprocess.check_output(GausPrefix + " < "  + f + ' > ' + f[:-3] + 'out', shell=True, timeout=86400)

        NCompleted += 1
        if IsGausCompleted(f[:-4] + '.out'):
            Completed.append(f[:-4] + '.out')
            print("Gaussian job " + str(NCompleted) + " of " + str(len(GausJobs)) + \
                  " completed.")
        else:
            print("Gaussian job terminated with an error. Continuing.")

    if NCompleted > 0:
        print(str(NCompleted) + " Gaussian jobs completed successfully.")

    return Completed


def WriteGausFile(Gausinp, conformer, atoms, charge, settings, type):

    f = open(Gausinp + '.com', 'w')
    if(settings.nProc > 1):
        f.write('%nprocshared=' + str(settings.nProc) + '\n')
    if settings.DFT == 'g':
        f.write('%mem=2000MB\n%chk='+Gausinp + '.chk\n')
    elif settings.DFT == 'z':
        f.write('%mem=1000MB\n%chk=' + Gausinp + '.chk\n')
    else:
        f.write('%mem=6000MB\n%chk='+Gausinp + '.chk\n')

    if type == 'nmr':
        f.write(NMRRoute(settings))
    elif type == 'e':
        f.write(ERoute(settings))
    elif type == 'opt':
        f.write(OptRoute(settings))

    f.write('\n'+Gausinp+'\n\n')
    f.write(str(charge) + ' 1\n')

    natom = 0

    for atom in conformer:
        f.write(atoms[natom] + '  ' + atom[0] + '  ' + atom[1] + '  ' +
                atom[2] + '\n')
        natom = natom + 1
    f.write('\n')

    f.close()


def NMRRoute(settings):

    route = '# ' + settings.nFunctional + '/' + settings.nBasisSet
    if (settings.nFunctional).lower() == 'm062x':
        route += ' int=ultrafine'

    route += ' nmr=giao'

    if settings.Solvent != '':
        route += ' scrf=(solvent=' + settings.Solvent + ')'

    route += '\n'

    return route


def ERoute(settings):

    route = '# ' + settings.eFunctional + '/' + settings.eBasisSet
    if (settings.eFunctional).lower() == 'm062x':
        route += ' int=ultrafine'

    if settings.Solvent != '':
        route += ' scrf=(solvent=' + settings.Solvent + ')'

    route += '\n'

    return route


def OptRoute(settings):

    route = '# ' + settings.oFunctional + '/' + settings.oBasisSet

    if (settings.oFunctional).lower() == 'm062x':
        route += ' int=ultrafine'

    route += ' Opt=(maxcycles=' + str(settings.MaxDFTOptCycles)
    if settings.CalcFC == True:
        route += ',CalcFC'
    if (settings.OptStepSize != 30):
        route += ',MaxStep=' + str(settings.OptStepSize)
    route += ')'

    if settings.Solvent != '':
        route += ' scrf=(solvent=' + settings.Solvent + ')'

    route += '\n'

    return route


def IsGausCompleted(f):
    Gfile = open(f, 'r')
    outp = Gfile.readlines()
    Gfile.close()
    if len(outp) < 10:
        return False
    if ("Normal termination" in outp[-1]) or (('termination' in '\n'.join(outp[-3:])) and ('l9999.exe' in '\n'.join(outp[-3:]))):
        return True
    else:
        return False


def IsGausConverged(f):
    Gfile = open(f, 'r')
    outp = Gfile.readlines()
    Gfile.close()
    ginp = '\n'.join(outp)
    if 'Stationary point found' in ginp:
        return True
    else:
        return False


#Read energy from e, if not present, then o, if not present, then nmr
def ReadEnergies(Isomers, settings):
    jobdir = os.getcwd()

    if 'e' in settings.Workflow:
        os.chdir('e')
    elif 'o' in settings.Workflow:
        os.chdir('opt')
    else:
        os.chdir('nmr')

    for i, iso in enumerate(Isomers):

        if 'e' in settings.Workflow:
            GOutpFiles = iso.EOutputFiles
        elif 'o' in settings.Workflow:
            GOutpFiles = iso.OptOutputFiles
        else:
            GOutpFiles = iso.NMROutputFiles

        DFTEnergies = []
        for GOutpFile in GOutpFiles:
            gausfile = open(GOutpFile, 'r')
            GOutp = gausfile.readlines()
            gausfile.close()

            for line in GOutp:
                if 'SCF Done:' in line:
                    start = line.index(') =')
                    end = line.index('A.U.')
                    energy = float(line[start + 4:end])

            DFTEnergies.append(energy)

        Isomers[i].DFTEnergies = DFTEnergies

        ####
        # Calculate rel. energies in kJ/mol
        minE = min(iso.DFTEnergies)

        relEs = []

        for e in iso.DFTEnergies:
            relEs.append((e - minE) * hartreeEnergy)

        Isomers[i].Energies = relEs

        populations = []

        # Calculate Boltzmann populations
        for e in relEs:
            populations.append(math.exp(-e * 1000 / (gasConstant * temperature)))

        q = sum(populations)

        for p in range(0, len(populations)):
            populations[p] = populations[p] / q

        Isomers[i].Populations = populations

        ####





    os.chdir(jobdir)
    return Isomers


def ReadShieldings(Isomers):

    jobdir = os.getcwd()
    os.chdir('nmr')

    for iso in Isomers:

        if len(iso.NMROutputFiles) < 1:
            print("Gaussian.py, ReadShieldings: No NMR DFT output" +
                  " files found, NMR data could not be read. Quitting.")
            quit()

        for GOutpFile in iso.NMROutputFiles:

            gausfile = open(GOutpFile, 'r')
            GOutp = gausfile.readlines()
            gausfile.close()

            index = 0
            shieldings = []
            labels = []

            # Find the NMR shielding calculation section
            while not 'Magnetic shielding' in GOutp[index]:
                index = index + 1

            # Read shielding constants and labels
            for line in GOutp[index:]:
                if 'Isotropic' in line:
                    data = [_f for _f in line.split(' ') if _f]
                    shieldings.append(float(data[4]))
                    labels.append(data[1] + data[0])

            print(GOutpFile,len(shieldings))

            iso.ConformerShieldings.append(shieldings)

        iso.ShieldingLabels = labels

    os.chdir(jobdir)

    return Isomers


def ReadGeometry(GOutpFile):

    gausfile = open(GOutpFile, 'r')
    GOutp = gausfile.readlines()
    gausfile.close()

    atoms = []
    coords = []
    gindex = -1

    #Find the last geometry section
    for index in range(len(GOutp)):
        if ('Input orientation:' in GOutp[index]) or ("Standard orientation:" in GOutp[index]):
            gindex = index + 5

    if gindex < 0:
        print('Error: No geometry found in file ' + GOutpFile)
        quit()

    #Read geometry
    for line in GOutp[gindex:]:
        if '--------------' in line:
            break
        else:
            data = [_f for _f in line[:-1].split(' ') if _f]
            atoms.append(GetAtomSymbol(int(data[1])))
            coords.append(data[3:])

    #return atoms, coords, charge

    return atoms, coords


def ReadGeometries(Isomers, settings):

    jobdir = os.getcwd()
    if ('o' in settings.Workflow):
        os.chdir('opt')

        for iso in Isomers:

            iso.DFTConformers = [[] for x in iso.OptOutputFiles]

            if len(iso.OptOutputFiles) < 1:
                print("Gaussian.py, ReadGeometries: No geometry optimisation output" +
                      " files found, geometries could not be read. Quitting.")
                quit()

            for num, GOutpFile in enumerate(iso.OptOutputFiles):

                atoms, coords = ReadGeometry(GOutpFile)

                iso.DFTConformers[num] = coords

            iso.Atoms = atoms
    else:
        os.chdir('nmr')

        for iso in Isomers:

            iso.DFTConformers = [[] for x in iso.NMROutputFiles]

            if len(iso.NMROutputFiles) < 1:
                print("Gaussian.py, ReadGeometries: No NMR DFT output" +
                      " files found, geometries could not be read. Quitting.")
                quit()

            for num, GOutpFile in enumerate(iso.NMROutputFiles):
                atoms, coords = ReadGeometry(GOutpFile)

                iso.DFTConformers[num] = coords

            iso.Atoms = atoms
    #return atoms, coords, charge
    os.chdir(jobdir)
    return Isomers


def GetAtomSymbol(AtomNum):
    Lookup = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg', 'Al', \
              'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V', 'Cr', 'Mn', 'Fe', 'Co', \
              'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se', 'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', \
              'Nb', 'Mo', 'Tc', 'Ru', 'Rh', 'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', \
              'Xe', 'Cs', 'Ba', 'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', \
              'Ho', 'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt', 'Au', \
              'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn']

    if AtomNum > 0 and AtomNum < len(Lookup):
        return Lookup[AtomNum-1]
    else:
        print("No such element with atomic number " + str(AtomNum))
        return 0



# -*- coding: utf-8 -*-
"""
Rewritten on Wed Jan 15 2020

@author: ke291

Contains all of the NWChem specific code for input generation and calculation
execution. Called by PyDP4.py.
"""

import glob
import os
import subprocess
import shutil

def SetupNMRCalcs(Isomers, settings):
    jobdir = os.getcwd()

    if not os.path.exists('nmr'):
        os.mkdir('nmr')
    os.chdir('nmr')

    for iso in Isomers:
        if iso.ExtCharge > -10:
            charge = iso.ExtCharge
        else:
            charge = iso.MMCharge

        if iso.DFTConformers == []:
            conformers = iso.Conformers
        else:
            conformers = iso.DFTConformers

        for num in range(0, len(conformers)):
            filename = iso.BaseName + 'nwinp' + str(num + 1).zfill(3)

            if os.path.exists(filename + '.nwo'):
                if IsNWChemCompleted(filename + '.nwo'):
                    iso.NMROutputFiles.append(filename + '.nwo')
                    continue
                else:
                    os.remove(filename + '.nwo')

            WriteNWChemFile(filename, conformers[num], iso.Atoms, charge, settings, 'nmr')
            iso.NMRInputFiles.append(filename + '.nw')

    os.chdir(jobdir)

    return Isomers


def SetupECalcs(Isomers, settings):

    jobdir = os.getcwd()

    if not os.path.exists('e'):
        os.mkdir('e')
    os.chdir('e')

    for iso in Isomers:
        if iso.ExtCharge > -10:
            charge = iso.ExtCharge
        else:
            charge = iso.MMCharge

        if iso.DFTConformers == []:
            conformers = iso.Conformers
        else:
            conformers = iso.DFTConformers

        for num in range(0, len(conformers)):
            filename = iso.BaseName + 'nwinp' + str(num + 1).zfill(3)

            if os.path.exists(filename + '.nwo'):
                if IsNWChemCompleted(filename + '.nwo'):
                    iso.EOutputFiles.append(filename + '.nwo')
                    continue
                else:
                    os.remove(filename + '.nwo')

            WriteNWChemFile(filename, conformers[num], iso.Atoms, charge, settings, 'e')
            iso.EInputFiles.append(filename + '.nw')

    os.chdir(jobdir)

    return Isomers


def SetupOptCalcs(Isomers, settings):
    jobdir = os.getcwd()

    if not os.path.exists('opt'):
        os.mkdir('opt')
    os.chdir('opt')

    for iso in Isomers:
        if iso.ExtCharge > -10:
            charge = iso.ExtCharge
        else:
            charge = iso.MMCharge

        if iso.DFTConformers == []:
            conformers = iso.Conformers
        else:
            conformers = iso.DFTConformers

        for num in range(0, len(conformers)):
            filename = iso.BaseName + 'nwinp' + str(num + 1).zfill(3)

            if os.path.exists(filename + '.nwo'):
                if IsNWChemCompleted(filename + '.nwo'):
                    iso.OptOutputFiles.append(filename + '.nwo')
                    continue
                else:
                    os.remove(filename + '.nwo')

            WriteNWChemFile(filename, conformers[num], iso.Atoms, charge, settings, 'opt')
            iso.OptInputFiles.append(filename + '.nw')

    os.chdir(jobdir)

    return Isomers


def Converged(Isomers):

    jobdir = os.getcwd()

    if not os.path.exists('opt'):
        os.chdir(jobdir)
        return False

    # insert code for convergence testing here

    os.chdir('opt')

    os.chdir(jobdir)
    return True


def RunNMRCalcs(Isomers, settings):

    print('\nRunning NWChem DFT NMR calculations locally...')

    jobdir = os.getcwd()
    os.chdir('nmr')

    NWJobs = []

    for iso in Isomers:
        print(iso.NMRInputFiles)
        NWJobs.extend([x for x in iso.NMRInputFiles if (x[:-3] + '.nwo') not in iso.NMROutputFiles])

    Completed = RunCalcs(NWJobs, settings)

    for iso in Isomers:
        iso.NMROutputFiles.extend([x[:-3] + '.nwo' for x in iso.NMRInputFiles if (x[:-3] + '.nwo') in Completed])

    os.chdir(jobdir)

    return Isomers


def GetPrerunNMRCalcs(Isomers):

    print('\nLooking for prerun NWChem DFT NMR files...')

    jobdir = os.getcwd()
    os.chdir('nmr')
    """
    for iso in Isomers:
        iso.NMRInputFiles = glob.glob(iso.BaseName + 'ginp*com')
        iso.NMROutputFiles.extend([x[:-4] + '.out' for x in iso.NMRInputFiles if IsGausCompleted(x[:-4] + '.out')])
    """
    print('NMR calc files:')
    print(', '.join([', '.join(x.NMROutputFiles) for x in Isomers]))

    os.chdir(jobdir)

    return Isomers


def RunECalcs(Isomers, settings):

    print('\nRunning NWChem DFT energy calculations locally...')

    jobdir = os.getcwd()
    os.chdir('e')

    NWJobs = []
    for iso in Isomers:
        print(iso.EInputFiles)
        NWJobs.extend([x for x in iso.EInputFiles if (x[:-3] + '.nwo') not in iso.EOutputFiles])

    print('To run: ' + str(NWJobs))
    Completed = RunCalcs(NWJobs, settings)

    for iso in Isomers:
        iso.EOutputFiles.extend([x[:-3] + '.nwo' for x in iso.EInputFiles if (x[:-3] + '.nwo') in Completed])

    os.chdir(jobdir)

    return Isomers


def GetPrerunECalcs(Isomers):

    print('\nLooking for prerun NWChem DFT energy calculation files...')

    jobdir = os.getcwd()
    os.chdir('e')

    """
    for iso in Isomers:
        iso.EInputFiles = glob.glob(iso.BaseName + 'ginp*com')
        iso.EOutputFiles.extend([x[:-4] + '.out' for x in iso.EInputFiles if IsGausCompleted(x[:-4] + '.out')])
    """
    print('Energy files:')
    print(', '.join([', '.join(x.EOutputFiles) for x in Isomers]))

    os.chdir(jobdir)

    return Isomers


def RunOptCalcs(Isomers, settings):

    print('\nRunning NWChem DFT geometry optimizations locally...')

    jobdir = os.getcwd()
    os.chdir('opt')

    NWJobs = []

    for iso in Isomers:
        print(iso.OptInputFiles)
        NWJobs.extend([x for x in iso.OptInputFiles if (x[:-3] + '.nwo') not in iso.OptOutputFiles])

    Completed = RunCalcs(NWJobs, settings)

    for iso in Isomers:
        iso.OptOutputFiles.extend([x[:-3] + '.nwo' for x in iso.OptInputFiles if (x[:-3] + '.nwo') in Completed])

    os.chdir(jobdir)

    return Isomers


def GetPrerunOptCalcs(Isomers):

    print('\nLooking for prerun NWChem DFT optimization files...')

    jobdir = os.getcwd()
    os.chdir('opt')
    """
    for iso in Isomers:
        iso.OptInputFiles = glob.glob(iso.BaseName + 'ginp*com')
        iso.OptOutputFiles.extend([x[:-4] + '.out' for x in iso.OptInputFiles if IsGausCompleted(x[:-4] + '.out')])
    """
    print('Opt files:')
    print(', '.join([', '.join(x.OptOutputFiles) for x in Isomers]))

    os.chdir(jobdir)

    return Isomers


def RunCalcs(NWJobs, settings):

    NCompleted = 0
    Completed = []
    NWChemPrefix = settings.NWChemPath

    if shutil.which(NWChemPrefix) is None:
        print('NWChem.py, RunCalcs:\n  Could not find NWChem executable at ' + NWChemPrefix)
        quit()

    for f in NWJobs:
        print(NWChemPrefix + ' ' + f + ' > ' + f[:-2] + 'nwo')
        outp = subprocess.check_output(NWChemPrefix + ' ' + f + ' > ' + f[:-2] +
                                       'nwo', shell=True)
        NCompleted += 1
        print("NWChem job " + str(NCompleted) + " of " + str(len(NWJobs)) + \
            " completed.")
        if IsNWChemCompleted(f[:-3] + '.nwo'):
            Completed.append(f[:-3] + '.nwo')
            print("NWChem job " + str(NCompleted) + " of " + str(len(NWJobs)) + \
                  " completed.")
        else:
            print("NWChem job terminated with an error. Continuing.")

    if NCompleted > 0:
        print(str(NCompleted) + " NWChem jobs completed successfully.")
    elif len(NWJobs) == 0:
        print("There were no jobs to run.")

    return Completed


def WriteNWChemFile(NWinp, conformer, atoms, charge, settings, type):

    f = open(NWinp + '.nw', 'w')
    f.write('memory stack 1500 mb heap 1500 mb global 3000 mb\n')
    if settings.DFT == 'w':
        f.write('scratch_dir /scratch/' + settings.user + '/' + NWinp + '\n')
    f.write('echo\n\nstart molecule\n\ntitle "' + NWinp + '"\n')
    f.write('echo\n\nstart\n\n')

    if settings.charge is not None:
        f.write('charge ' + str(settings.charge) + '\n\n')
    else:
        f.write('charge ' + str(charge) + '\n\n')

    f.write('geometry units angstroms print xyz autosym\n')

    natom = 0
    for atom in conformer:
        f.write('  ' + atoms[natom] + ' ' + atom[0] + ' ' + atom[1] + ' ' +
                atom[2] + '\n')
        natom = natom + 1

    basis = settings.nBasisSet
    if basis.lower() == '6-31g(d,p)':
        basis = '6-31g**'
    elif basis.lower() == '6-311g(d)':
        basis = '6-311g*'

    f.write('end\n\nbasis\n  * library ' + basis + '\nend\n\n')
    if settings.Solvent != "":
        GausSolvents = ['chloroform', 'dimethylsulfoxide', 'benzene', 'methanol', 'pyridine', 'acetone']
        NWSolvents = ['chcl3', 'dmso', 'benzene', 'methanol', 'pyridine', 'acetone']

        if settings.Solvent in GausSolvents:
            solvent = NWSolvents[GausSolvents.index(settings.Solvent)]
        else:
            solvent = settings.Solvent

        f.write('cosmo\n  do_cosmo_smd true\n  solvent ' + solvent + '\n')
        f.write('end\n\n')

    if type == 'nmr':
        f.write(NMRSuffix(settings))
    elif type == 'e':
        f.write(ESuffix(settings))
    elif type == 'opt':
        f.write(OptSuffix(settings))

    f.close()


def NMRSuffix(settings):
    suffix = ''
    if (settings.nFunctional).lower() == 'b3lyp':
        suffix += 'dft\n  xc b3lyp\n  mult 1\nend\n\n'
    elif (settings.nFunctional).lower() == 'm062x' or \
            (settings.nFunctional).lower() == 'm06-2x':
        suffix +=  'dft\n  xc m06-2x\n  mult 1\nend\n\n'
    elif (settings.nFunctional).lower() == 'mpw1pw91':
        suffix += 'dft\n  xc mpw91 0.75 HFexch 0.25 perdew91\n  mult 1\nend\n\n'
    else:
        suffix += 'dft\n  xc ' + settings.nFunctional + '\n  mult 1\nend\n\n'
    suffix += 'task dft energy\n\nproperty\n  shielding\nend\ntask dft property\n'

    return suffix


def ESuffix(settings):
    suffix = ''
    if (settings.nFunctional).lower() == 'b3lyp':
        suffix += 'dft\n  xc b3lyp\n  mult 1\nend\n\n'
    elif (settings.nFunctional).lower() == 'm062x' or \
            (settings.nFunctional).lower() == 'm06-2x':
        suffix +=  'dft\n  xc m06-2x\n  mult 1\nend\n\n'
    elif (settings.nFunctional).lower() == 'mpw1pw91':
        suffix += 'dft\n  xc mpw91 0.75 HFexch 0.25 perdew91\n  mult 1\nend\n\n'
    else:
        suffix += 'dft\n  xc ' + settings.nFunctional + '\n  mult 1\nend\n\n'
    suffix += 'task dft energy\n'

    return suffix


def OptSuffix(settings):
    suffix = 'driver\n  maxiter ' + str(settings.MaxDFTOptCycles) + '\nend\n\n'
    if (settings.oFunctional).lower() == 'b3lyp':
        suffix += 'dft\n  xc b3lyp\n  mult 1\nend\n\n'
        suffix += 'task dft optimize\n\n'
    elif (settings.oFunctional).lower() == 'm062x' or (settings.oFunctional).lower() == 'm06-2x':
        suffix += 'dft\n  xc m06-2x\n  mult 1\nend\n\n'
        suffix += 'task dft optimize\n\n'

    return suffix


def IsNWChemCompleted(f):
    Nfile = open(f, 'r')
    outp = Nfile.readlines()
    Nfile.close()
    outp = "".join(outp)
    if "AUTHORS" in outp:
        return True
    else:
        return False


def IsNWChemConverged(f):
    Nfile = open(f, 'r')
    outp = Nfile.readlines()
    Nfile.close()
    outp = "".join(outp)
    if 'Optimization converged' in outp:
        return True
    else:
        return False


#Read energy from e, if not present, then o, if not present, then nmr
def ReadEnergies(Isomers, settings):
    jobdir = os.getcwd()

    if 'e' in settings.Workflow:
        os.chdir('e')
    elif 'o' in settings.Workflow:
        os.chdir('opt')
    else:
        os.chdir('nmr')

    for i, iso in enumerate(Isomers):

        if 'e' in settings.Workflow:
            NWOutpFiles = iso.EOutputFiles
        elif 'o' in settings.Workflow:
            NWOutpFiles = iso.OptOutputFiles
        else:
            NWOutpFiles = iso.NMROutputFiles

        DFTEnergies = []
        for NWOutpFile in NWOutpFiles:
            nwfile = open(NWOutpFile, 'r')
            NWOutp = nwfile.readlines()
            nwfile.close()

            for line in NWOutp:
                if 'Total DFT energy' in line:
                    start = line.index('Total')
                    energy = float(line[start + 19:])

            DFTEnergies.append(energy)

        Isomers[i].DFTEnergies = DFTEnergies

    os.chdir(jobdir)
    return Isomers


def ReadShieldings(Isomers):
    jobdir = os.getcwd()
    os.chdir('nmr')

    for iso in Isomers:

        if len(iso.NMROutputFiles) < 1:
            print("NWChem.py, ReadShieldings: No NMR DFT output" +
                  " files found, NMR data could not be read. Quitting.")
            quit()

        for NWOutpFile in iso.NMROutputFiles:
            nwfile = open(NWOutpFile, 'r')
            NWOutp = nwfile.readlines()
            nwfile.close()

            index = 0
            shieldings = []
            labels = []

            # Find the NMR shielding calculation section
            while not 'Chemical Shielding' in NWOutp[index]:
                index = index + 1

            # Read shielding constants and labels
            for line in NWOutp[index:]:
                if 'isotropic' in line:
                    start = line.index('isotropic')
                    shieldings.append(float(line[start + 13:]))
                if 'Atom' in line:
                    start = line.index('Atom')
                    labels.append(line[start + 12] + line[start + 7:start + 10].strip())

            print(NWOutpFile,len(shieldings))

            iso.ConformerShieldings.append(shieldings)

        iso.ShieldingLabels = labels

    os.chdir(jobdir)

    return Isomers


def ReadGeometry(NWOutpFile):

    nwfile = open(NWOutpFile, 'r')
    NWOutp = nwfile.readlines()
    nwfile.close()

    atoms = []
    coords = []
    gindex = -1

    # Find the last geometry section
    for index in range(len(NWOutp)):
        if ('Geometry "geometry" -> "geometry"' in NWOutp[index]):
            gindex = index + 7

    if gindex < 0:
        print('Error: No geometry found in file ' + NWOutpFile)
        quit()

    # Read geometry
    for line in NWOutp[gindex:]:
        if len(line) < 2:
            break
        else:
            data = [_f for _f in line[:-1].split(' ') if _f]
            atoms.append(data[1])
            coords.append(data[3:])

    return atoms, coords


def ReadGeometries(Isomers, settings):

    jobdir = os.getcwd()
    if ('o' in settings.Workflow):
        os.chdir('opt')

        for iso in Isomers:

            iso.DFTConformers = [[] for x in iso.OptOutputFiles]

            if len(iso.OptOutputFiles) < 1:
                print("NWChem.py, ReadGeometries: No geometry optimisation output" +
                      " files found, geometries could not be read. Quitting.")
                quit()

            for num, NWOutpFile in enumerate(iso.OptOutputFiles):

                atoms, coords = ReadGeometry(NWOutpFile)

                iso.DFTConformers[num] = coords

            iso.Atoms = atoms
    else:
        os.chdir('nmr')

        for iso in Isomers:

            iso.DFTConformers = [[] for x in iso.NMROutputFiles]

            if len(iso.NMROutputFiles) < 1:
                print("NWChem.py, ReadGeometries: No geometry optimisation output" +
                      " files found, geometries could not be read. Quitting.")
                quit()

            for num, NWOutpFile in enumerate(iso.NMROutputFiles):
                atoms, coords = ReadGeometry(NWOutpFile)

                iso.DFTConformers[num] = coords

            iso.Atoms = atoms
    #return atoms, coords, charge
    os.chdir(jobdir)
    return Isomers




import itertools
import json
import numpy as np
from lmfit import Parameters
from pathlib import Path

from ..helper_functions import lorentzian


def solvent_removal(simulated_y_data, spectral_xdata_ppm, solvent, uc, picked_peaks):

    solvent_file = (Path(__file__).parent / "solvents.json").resolve()

    with open(solvent_file) as f:
        solvent_dict = json.load(f)

    if solvent in solvent_dict:
        solvent_data = solvent_dict[solvent]
    else:
        solvent_data = []

        exp_ppm = []
        Jv = [[]]

    # remove all solvent peaks

    removed = []

    for peak in solvent_data:

        p, J = peak["exp_ppm"], peak["Jv"]

        exp = uc(p, "ppm")

        region = np.arange(exp - 1000, exp + 1000)

        peak_region = []

        for peak in picked_peaks:

            if (peak > exp - 1000) & (peak < exp + 1000):
                peak_region.append(peak)

        # simulate solvent curve

        # find peak centre

        if region[0] + region[-1] & 1:
            centre = int((region[0] + region[-1] + 1) / 2)
        else:
            centre = int((region[0] + region[-1]) / 2)

        centre = uc.ppm(centre)

        params, peak_vector, amp_vector, y = first_order_peak(
            centre, J, np.array(region), 1, uc, 1
        )

        # use simulated curve in convolution

        convolved_y = np.convolve(simulated_y_data[region], y, "same")

        mxpoint = np.argmax(convolved_y)

        mxppm = uc.ppm(region[mxpoint])

        # simulate peak in new position

        params, fit_s_peaks, amp_vector, fit_s_y = first_order_peak(
            mxppm, J, np.array(region), 1, uc, 1
        )

        # find average of fitted peaks for referencing:

        av = sum(fit_s_peaks) / len(fit_s_peaks)

        avppm = uc.ppm(av)

        spectral_xdata_ppm -= avppm - p

        to_remove = []

        # find picked peaks closest to the "fitted" solvent multiplet

        for peak in fit_s_peaks:

            i = np.abs(np.array(picked_peaks) - peak).argmin()

            to_remove.append(i)

        removed.extend([picked_peaks[i] for i in to_remove])

        to_remove = sorted(list(set(to_remove)), reverse=True)

        for peak in to_remove:
            picked_peaks.pop(peak)

    removed = np.array(removed)

    return picked_peaks, removed


def first_order_peak(start_ppm, J_vals, x_data, corr_distance, uc, m):

    # new first order peak generator using the method presented in Hoye paper

    start = uc(str(start_ppm) + "ppm")

    start_Hz = uc.hz(start)

    J_vals = np.array(J_vals)

    if len(J_vals) > 0:

        peaks = np.zeros((2 * m + 1) ** len(J_vals))

        if m == 0.5:

            l = [1, -1]

        if m == 1:

            l = [1, 0, -1]

        # signvector generator

        signvectors = itertools.product(l, repeat=len(J_vals))

        shifts = []

        for ind, sv in enumerate(signvectors):
            shift = J_vals * sv
            shift = start_Hz + 0.5 * (np.sum(shift))
            peaks[ind] = shift

        peaks = np.sort(peaks)

        peak_vector = np.array(sorted(list(set(peaks)), reverse=True))

        amp_vector = np.zeros(len(peak_vector))

        for peak in peaks:
            index = np.where(peak_vector == peak)
            amp_vector[index] += 1

        pv = []

        for index, peak in enumerate(peak_vector):

            pv.append(uc(peak, "Hz"))

        peak_vector = pv

    else:
        peak_vector = start

    split_params = Parameters()

    for index, peak in enumerate(peak_vector):
        split_params.add("amp" + str(peak), value=amp_vector[index])
        split_params.add("pos" + str(peak), value=peak)
        split_params.add("width" + str(peak), value=2 * corr_distance)

    y = lorenz_curves(split_params, x_data, peak_vector)

    y = y / np.max(y)

    return split_params, peak_vector, amp_vector, y


def lorenz_curves(params, x, picked_points):

    y = np.zeros(len(x))
    for peak in picked_points:
        y += lorentzian(
            x,
            params["width" + str(peak)],
            params["pos" + str(peak)],
            params["amp" + str(peak)],
        )
    return y



#!/usr/bin/env python

# -*- coding: utf-8 -*-
"""
Created on Wed Apr 19 15:56:54 2019

@author: ke291

Contains all of the specific code for running Gaussian jobs on Darwin.
A lot of code is reused from Gaussian.py. Called by PyDP4.py.
"""

import subprocess
import os
import time
import shutil
import math

import dp5.old_files.Gaussian as Gaussian

MaxConcurrentJobs = 160

SetupNMRCalcs = Gaussian.SetupNMRCalcs

SetupECalcs = Gaussian.SetupECalcs

SetupOptCalcs = Gaussian.SetupOptCalcs

ReadShieldings = Gaussian.ReadShieldings

ReadEnergies = Gaussian.ReadEnergies

ReadGeometries = Gaussian.ReadGeometries

IsGausCompleted = Gaussian.IsGausCompleted

Converged = Gaussian.Converged


def RunNMRCalcs(Isomers, settings):
    print('\nRunning Gaussian NMR calculations on Darwin...')

    # Run Gaussian jobs on CSD3 cluster in folder named after date and time
    # Split in batches, if needed

    jobdir = os.getcwd()
    os.chdir('nmr')

    GausJobs = []

    for iso in Isomers:
        GausJobs.extend([x for x in iso.NMRInputFiles if (
            x[:-4] + '.out') not in iso.NMROutputFiles])

    Completed = RunCalcs(GausJobs, settings)

    for iso in Isomers:
        iso.NMROutputFiles.extend(
            [x[:-4] + '.out' for x in iso.NMRInputFiles if (x[:-4] + '.out') in Completed])

    os.chdir(jobdir)

    return Isomers


def RunECalcs(Isomers, settings):
    print('\nRunning Gaussian DFT energy calculations on Darwin...')

    jobdir = os.getcwd()
    os.chdir('e')

    GausJobs = []

    for iso in Isomers:
        GausJobs.extend([x for x in iso.EInputFiles if (
            x[:-4] + '.out') not in iso.EOutputFiles])

    Completed = RunCalcs(GausJobs, settings)

    for iso in Isomers:
        iso.EOutputFiles.extend(
            [x[:-4] + '.out' for x in iso.EInputFiles if (x[:-4] + '.out') in Completed])

    os.chdir(jobdir)

    return Isomers


def RunOptCalcs(Isomers, settings):
    print('\nRunning Gaussian DFT geometry optimizations on Darwin...')

    jobdir = os.getcwd()
    os.chdir('opt')

    GausJobs = []

    for iso in Isomers:
        GausJobs.extend([x for x in iso.OptInputFiles if (
            x[:-4] + '.out') not in iso.OptOutputFiles])

    Completed = RunCalcs(GausJobs, settings)

    for iso in Isomers:
        iso.OptOutputFiles.extend(
            [x[:-4] + '.out' for x in iso.OptInputFiles if (x[:-4] + '.out') in Completed])

    os.chdir(jobdir)

    return Isomers


def RunCalcs(GausJobs, settings):

    MaxCon = MaxConcurrentJobs

    if len(GausJobs) < MaxCon:
        if len(GausJobs) > 0:
            RunBatchOnDarwin(0, GausJobs, settings)
    else:
        if len(GausJobs) > 0:
            print("The DFT calculations will be done in " +
                  str(math.ceil(len(GausJobs) / MaxCon)) + " batches")
            i = 0
            while (i + 1) * MaxCon < len(GausJobs):
                print("Starting batch nr " + str(i + 1))
                RunBatchOnDarwin(
                    str(i + 1), GausJobs[(i * MaxCon):((i + 1) * MaxCon)], settings)
                i += 1
            print("Starting batch nr " + str(i + 1))
            RunBatchOnDarwin(str(i + 1), GausJobs[(i * MaxCon):], settings)

    NCompleted = 0
    Completed = []

    for f in GausJobs:
        if IsGausCompleted(f[:-4] + '.out'):
            Completed.append(f[:-4] + '.out')
            NCompleted += 1

    print(str(NCompleted) + "Gaussian jobs of " + str(len(GausJobs)) +
          " completed successfully.")

    return Completed


def RunBatchOnDarwin(findex, GausJobs, settings):

    if findex == 0:
        folder = settings.StartTime + settings.Title
    else:
        folder = settings.StartTime + findex + settings.Title

    scrfolder = settings.StartTime + settings.Title

    # Check that results folder does not exist, create job folder on darwin
    outp = subprocess.Popen(['ssh', 'darwin', 'ls'],
                            stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]
    print("Results folder: " + folder)

    if folder in outp.decode():
        print("Results folder exists on Darwin, choose another folder name.")
        quit()

    outp = subprocess.Popen(['ssh', 'darwin', 'mkdir', folder],
                            stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]

    # Check that scratch directory does not exist, create job folder on darwin
    outp = subprocess.Popen(['ssh', 'darwin', 'ls ' + settings.DarwinScrDir],
                            stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]
    print("Scratch directory: " + settings.DarwinScrDir + scrfolder)

    if folder in outp.decode():
        print("Scratch folder exists on Darwin, choose another folder name.")
        quit()

    outp = subprocess.Popen(['ssh', 'darwin', 'mkdir', settings.DarwinScrDir + scrfolder],
                            stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]

    # Write the slurm scripts
    SubFiles = WriteDarwinScripts(GausJobs, settings, scrfolder)

    print(str(len(SubFiles)) + ' slurm scripts generated')

    # Upload .com files and slurm files to directory
    print("Uploading files to darwin...")
    for f in GausJobs:
        outp = subprocess.Popen(['scp', f,
                                 'darwin:/home/' + settings.user + '/' + folder],
                                stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]

    for f in SubFiles:

        outp = subprocess.Popen(['scp', f,
                                 'darwin:/home/' + settings.user + '/' + folder],
                                stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]

    print(str(len(GausJobs)) + ' .com and ' + str(len(SubFiles)) +
          ' slurm files uploaded to darwin')

    fullfolder = '/home/' + settings.user + '/' + folder
    JobIDs = []

    # Launch the calculations
    for f in SubFiles:
        outp = subprocess.Popen(['ssh', 'darwin', 'cd ' + fullfolder + ';sbatch', f],
                                stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]
        status = outp.decode().split('\n')[-2]
        print(status)
        JobIDs.append(status.split('job ')[1])

    print(str(len(SubFiles)) + ' jobs submitted to the queue on darwin ' +
          'containing ' + str(len(GausJobs)) + ' Gaussian jobs')

    time.sleep(60)

    OldQRes = CheckDarwinQueue(JobIDs, settings)

    while OldQRes[0] < 0:
        OldQRes = CheckDarwinQueue(JobIDs, settings)
        time.sleep(60)

    print('Pending: ' + str(OldQRes[0]) + ', Running: ' +
          str(OldQRes[1]) + ', Not in queue: ' + str(OldQRes[2]))

    Jobs2Complete = list(GausJobs)
    n2complete = len(Jobs2Complete)

    # Check and report on the progress of calculations
    while len(Jobs2Complete) > 0:
        JobFinished = IsDarwinGComplete(Jobs2Complete, folder, settings)

        Jobs2Complete[:] = [job for job in Jobs2Complete if
                            not JobFinished[job[:-3] + 'out']]
        if n2complete != len(Jobs2Complete):
            n2complete = len(Jobs2Complete)
            print(str(n2complete) + " Gaussian jobs remaining.")

        QRes = CheckDarwinQueue(JobIDs, settings)
        if QRes != OldQRes:
            if QRes[0] < 0:
                QRes = OldQRes
            else:
                OldQRes = QRes
                print('Darwin queue:')
                print('Pending: ' + str(OldQRes[0]) + ', Running: ' +
                      str(OldQRes[1]) + ', Not in queue: ' + str(OldQRes[2]))

        if (QRes[2] == len(JobIDs)) and (QRes[0] >= 0):
            # check each gaussian file to ascertain the status of individual gaus jobs
            print('No jobs left in Darwin queue')
            break

        time.sleep(180)

    # When done, copy the results back
    print("\nCopying the output files back to localhost...")
    print('scp darwin:' + fullfolder + '/*.out ' + os.getcwd() + '/')

    outp = subprocess.Popen(['scp', 'darwin:' + fullfolder + '/*.out',
                             os.getcwd() + '/'],
                            stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]

    print("\nDeleting checkpoint files...")
    print('ssh darwin rm ' + fullfolder + '/*.chk')
    outp = subprocess.Popen(['ssh', 'darwin', 'rm', fullfolder + '/*.chk'],
                            stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]

    fullscrfolder = settings.DarwinScrDir + scrfolder
    print("\nDeleting scratch folder...")
    print('ssh darwin rm -r ' + fullscrfolder)

    outp = subprocess.Popen(['ssh', 'darwin', 'rm -r', fullscrfolder],
                            stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]


def WriteDarwinScripts(GausJobs, settings, scrfolder):

    SubFiles = []
    AdjNodeSize = int(math.floor(settings.DarwinNodeSize/settings.nProc))

    # If the jobs exactly fill the node, just write the submission script
    if len(GausJobs) == AdjNodeSize:
        SubFiles.append(WriteSlurm(GausJobs, settings, scrfolder))

    # If the jobs do not fill the node, increase the processor count to fill it
    elif len(GausJobs) < AdjNodeSize:
        NewNProc = int(math.floor(settings.DarwinNodeSize / len(GausJobs)))
        SubFiles.append(WriteSlurm(GausJobs, settings,
                        scrfolder, nProc=NewNProc))
        print("Jobs don't fill the Darwin node, nproc increased to " + str(NewNProc))
        for j, GausJob in enumerate(GausJobs):
            line = '%nprocshared=' + str(NewNProc) + '\n'
            ReplaceLine(GausJob, 0, line)

    # If the jobs more than fill the node, submit as several jobs
    else:
        print("The Gaussian calculations will be submitted as " +
              str(math.ceil(len(GausJobs)/AdjNodeSize)) +
              " jobs")
        i = 0
        while (i+1)*AdjNodeSize < len(GausJobs):
            PartGausJobs = list(GausJobs[(i*AdjNodeSize):((i+1)*AdjNodeSize)])
            print("Writing script nr " + str(i+1))
            SubFiles.append(WriteSlurm(
                PartGausJobs, settings, scrfolder, str(i+1)))

            i += 1

        PartGausJobs = list(GausJobs[(i*AdjNodeSize):])

        # if the last few jobs do not fill the node, increase the processor count to fill it
        if len(PartGausJobs) < AdjNodeSize:
            NewNProc = int(math.floor(
                settings.DarwinNodeSize / len(PartGausJobs)))
            print(
                "Jobs don't fill the last Darwin node, nproc increased to " + str(NewNProc))
            print("Writing script nr " + str(i + 1))
            SubFiles.append(WriteSlurm(PartGausJobs, settings,
                            scrfolder, str(i+1), nProc=NewNProc))
            for j, GausJob in enumerate(PartGausJobs):
                line = '%nprocshared=' + str(NewNProc) + '\n'
                ReplaceLine(GausJob, 0, line)
        else:
            print("Writing script nr " + str(i + 1))
            SubFiles.append(WriteSlurm(
                PartGausJobs, settings, scrfolder, str(i+1)))

    return SubFiles


def WriteSlurm(GausJobs, settings, scrfolder, index='', nProc=-1):

    if nProc == -1:
        nProc = settings.nProc

    cwd = os.getcwd()
    filename = settings.Title + 'slurm' + index

    shutil.copyfile(settings.ScriptDir + '/Defaultslurm',
                    cwd + '/' + filename)
    slurmf = open(filename, 'r+')
    slurm = slurmf.readlines()
    slurm[12] = '#SBATCH -J ' + settings.Title + '\n'
    slurm[14] = '#SBATCH -A ' + settings.project + '\n'
    slurm[19] = '#SBATCH --ntasks=' + str(len(GausJobs)*nProc) + '\n'
    slurm[21] = '#SBATCH --time=' + format(settings.TimeLimit, "02") +\
        ':00:00\n'

    slurm[61] = 'export GAUSS_SCRDIR=' + \
        settings.DarwinScrDir + scrfolder + '\n'

    for f in GausJobs:
        slurm.append('srun --exclusive -n1 -c' + str(nProc) + ' $application < ' + f[:-3] +
                     'com > ' + f[:-3] + 'out 2> error &\n')

    slurm.append('wait\n')

    slurmf.truncate(0)
    slurmf.seek(0)
    slurmf.writelines(slurm)

    return filename


def IsDarwinGComplete(GausJobs, folder, settings):

    path = '/home/' + settings.user + '/' + folder + '/'
    results = {}

    for f in GausJobs:
        outp = subprocess.Popen(['ssh', 'darwin', 'cat', path + f[:-3] + 'out'],
                                stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]

        if "Normal termination" in outp[-90:].decode():
            results[f[:-3] + 'out'] = True
        else:
            results[f[:-3] + 'out'] = False

    return results


def CheckDarwinQueue(JobIDs, settings):

    outp = subprocess.Popen(['ssh', 'darwin', 'squeue', '-u ' + settings.user],
                            stderr=subprocess.STDOUT, stdout=subprocess.PIPE).communicate()[0]
    outp = outp.decode().split('\n')
    QStart = -1000
    for i, line in enumerate(outp):
        if 'JOBID' in line:
            QStart = i+1
            break

    if QStart < 0:
        return -100, -100, -100

    QueueReport = outp[QStart:-1]
    JobStats = []

    for job in JobIDs:
        status = ''
        for i, line in enumerate(QueueReport):
            if job in line:
                status = list(filter(None, line.split(' ')))[4]
        JobStats.append(status)

    Pending = JobStats.count('PD')
    Running = JobStats.count('R')
    NotInQueue = JobStats.count('')

    return Pending, Running, NotInQueue


def ReplaceLine(File, LineN, Line):
    gausf = open(File, 'r+')
    gauslines = gausf.readlines()
    gauslines[LineN] = Line
    gausf.truncate(0)
    gausf.seek(0)
    gausf.writelines(gauslines)
    gausf.close()



import multiprocessing as mp
import copy

import numpy as np
from lmfit import Minimizer, Parameters

from dp5.nmr_processing.proton.pearson7 import p7sim, p7residual


def BIC_minimisation_region_full(
    ind1, uc, peak_regions, grouped_peaks, total_spectral_ydata, corr_distance, std
):

    ################################################################################################################
    # initialise process
    ################################################################################################################

    # print("minimising region " + str(ind1) + " of " + str(len(peak_regions)))

    BIC_param = 15

    region = np.array(peak_regions[ind1])

    region_y = total_spectral_ydata[region]

    fit_y = np.zeros(len(region_y))

    copy_peaks = np.array(grouped_peaks[ind1])

    params = Parameters()

    fitted_peaks = []

    ttotal = 0

    ################################################################################################################
    # build initial model
    ################################################################################################################

    # params.add('vregion' + str(ind1), value=2.5, max=5, min=1)

    params.add("vregion" + str(ind1), value=0.5, max=1, min=0)

    distance = uc(0, "hz") - uc(5, "hz")

    std_upper = uc(0, "hz") - uc(1, "hz")
    av_std = uc(0, "hz") - uc(0.2, "hz")
    std_lower = uc(0, "hz") - uc(0.1, "hz")

    # build model

    while len(copy_peaks) > 0:
        # pick peak that is furthest from fitted data:

        diff_array = region_y - fit_y

        ind2 = np.argmax(diff_array[copy_peaks - region[0]])

        maxpeak = copy_peaks[ind2]

        copy_peaks = np.delete(copy_peaks, ind2)

        # only allow params < distance away vary at a time

        # add new params

        fitted_peaks.append(maxpeak)

        fitted_peaks = sorted(fitted_peaks)

        params.add(
            "A" + str(maxpeak),
            value=total_spectral_ydata[maxpeak],
            min=0,
            max=1,
            vary=True,
        )

        # params.add('std' + str(maxpeak), value=av_std, vary=True, min = std_lower,
        #               max = std_upper)

        params.add("std" + str(maxpeak), value=av_std, vary=True)

        params.add(
            "mu" + str(maxpeak),
            value=maxpeak,
            vary=True,
            min=maxpeak - 4 * corr_distance,
            max=maxpeak + 4 * corr_distance,
        )

        # adjust amplitudes and widths of the current model

    initial_y = p7sim(params, region, fitted_peaks, ind1)

    inty = np.sum(region_y[region_y > 0])

    intmodel = np.sum(initial_y)

    # check the region can be optimised this way

    # find peak with max amplitude

    maxamp = 0

    for peak in fitted_peaks:
        amp = params["A" + str(peak)]
        if amp > maxamp:
            maxamp = copy.copy(amp)

    maxintegral = maxamp * len(region)

    if maxintegral > inty:

        # set initial conditions

        while (intmodel / inty < 0.99) or (intmodel / inty > 1.01):

            for f in fitted_peaks:
                params["std" + str(f)].set(
                    value=params["std" + str(f)] * inty / intmodel
                )

            initial_y = p7sim(params, region, fitted_peaks, ind1)

            for f in fitted_peaks:
                params["A" + str(f)].set(
                    value=params["A" + str(f)]
                    * region_y[int(params["mu" + str(f)]) - region[0]]
                    / (initial_y[f - region[0]])
                )

            initial_y = p7sim(params, region, fitted_peaks, ind1)

            intmodel = np.sum(initial_y)

    # print('built model region ' + str(ind1))

    ################################################################################################################
    # now relax all params
    ################################################################################################################

    # allow all params to vary

    params["vregion" + str(ind1)].set(vary=True)

    for peak in fitted_peaks:
        params["A" + str(peak)].set(
            vary=False,
            min=max(0, params["A" + str(peak)] - 0.01),
            max=min(params["A" + str(peak)] + 0.01, 1),
        )
        params["mu" + str(peak)].set(vary=False)
        params["std" + str(peak)].set(
            vary=False,
            min=min(std_lower, params["std" + str(peak)] - av_std),
            max=max(params["std" + str(peak)] + av_std, std_upper),
        )

    out = Minimizer(
        p7residual, params, fcn_args=(region, fitted_peaks, region_y, ind1, False)
    )

    results = out.minimize()

    params = results.params

    # print('relaxed params region ' + str(ind1))

    ################################################################################################################
    # now remove peaks in turn
    ################################################################################################################

    trial_y = p7sim(params, region, fitted_peaks, ind1)

    trial_peaks = np.array(fitted_peaks)

    amps = []

    for peak in trial_peaks:
        amps.append(params["A" + str(peak)])

    r = trial_y - region_y

    chi2 = r**2

    N = len(chi2)

    BIC = N * np.log(np.sum(chi2) / N) + np.log(N) * (3 * len(fitted_peaks) + 2)

    while len(trial_peaks) > 0:

        new_params = copy.copy(params)

        # find peak with smallest amp

        minpeak = trial_peaks[np.argmin(amps)]

        # remove this peak from the set left to try

        trial_peaks = np.delete(trial_peaks, np.argmin(amps))
        amps = np.delete(amps, np.argmin(amps))

        # remove this peak from the trial peaks list and the trial params

        new_params.__delitem__("A" + str(minpeak))
        new_params.__delitem__("mu" + str(minpeak))
        new_params.__delitem__("std" + str(minpeak))

        new_fitted_peaks = np.delete(fitted_peaks, np.where(fitted_peaks == minpeak))

        # simulate data with one fewer peak

        new_trial_y = p7sim(new_params, region, new_fitted_peaks, ind1)

        r = new_trial_y - region_y

        chi2 = np.sum(r**2)

        N = len(new_trial_y)

        new_BIC = N * np.log(chi2 / N) + np.log(N) * (3 * len(new_fitted_peaks) + 2)

        # if the fit is significantly better remove this peak

        if new_BIC < BIC - BIC_param:
            fitted_peaks = copy.copy(new_fitted_peaks)

            params = copy.copy(new_params)

            BIC = copy.copy(new_BIC)

    fitted_peaks = sorted(fitted_peaks)

    fit_y = p7sim(params, region, fitted_peaks, ind1)

    ################################################################################################################

    return fitted_peaks, params, fit_y


def multiproc_BIC_minimisation(
    peak_regions, grouped_peaks, total_spectral_ydata, corr_distance, uc, std
):
    maxproc = mp.cpu_count()

    pool = mp.Pool(maxproc)

    new_grouped_peaks = [[] for i in range(len(peak_regions))]
    new_grouped_params = [[] for i in range(len(peak_regions))]
    new_sim_y = [[] for i in range(len(peak_regions))]

    # order regions by size to efficiently fill cores

    region_lengths = np.array([len(g) for g in grouped_peaks])

    sorted_regions = np.argsort(region_lengths)[::-1]

    res = [[] for i in peak_regions]

    # write output files

    for ind1 in range(len(peak_regions)):
        res[ind1] = pool.apply_async(
            BIC_minimisation_region_full,
            [
                ind1,
                uc,
                peak_regions,
                grouped_peaks,
                total_spectral_ydata,
                corr_distance,
                std,
            ],
        )

    for ind1 in sorted_regions:
        new_grouped_peaks[ind1], new_grouped_params[ind1], new_sim_y[ind1] = res[
            ind1
        ].get()

    # unpack the parameters and split groups

    final_grouped_peaks = []
    final_peaks = []

    total_params = Parameters()

    new_peaks = []

    dist_hz = uc(0, "Hz") - uc(20, "Hz")

    newgroupind = 0
    oldgroupind = 0

    for group in new_grouped_peaks:

        group = sorted(group)

        if len(group) > 0:
            final_grouped_peaks.append([])
            total_params.add(
                "vregion" + str(newgroupind),
                value=new_grouped_params[oldgroupind]["vregion" + str(oldgroupind)],
            )

            final_peaks.extend(group)

        for ind2, peak in enumerate(group):

            # check if the group should be split

            if ind2 > 0:

                # if there is a gap of more than 20Hz split the group

                if peak > group[ind2 - 1] + dist_hz:
                    # track the numer of splits included to ensure v parameter is added to the correct group each time

                    newgroupind += 1

                    # if a split occures add a new v parameter for the new group

                    total_params.add(
                        "vregion" + str(newgroupind),
                        value=new_grouped_params[oldgroupind][
                            "vregion" + str(oldgroupind)
                        ],
                    )

                    # allow peaks to be added to the new group
                    final_grouped_peaks.append([])

            # finally append the peak
            final_grouped_peaks[-1].append(peak)
            total_params.add(
                "A" + str(peak), value=new_grouped_params[oldgroupind]["A" + str(peak)]
            )
            total_params.add(
                "std" + str(peak),
                value=new_grouped_params[oldgroupind]["std" + str(peak)],
            )
            total_params.add(
                "mu" + str(peak),
                value=new_grouped_params[oldgroupind]["mu" + str(peak)],
                vary=False,
            )

        if len(group) > 0:
            newgroupind += 1

        oldgroupind += 1

    # draw regions between midpoints of groups
    final_peak_regions = np.empty(len(final_grouped_peaks), dtype=object)
    for ind4, group in enumerate(final_grouped_peaks):

        if ind4 == 0:
            lower_point = 0
            higher_point = int((group[-1] + final_grouped_peaks[ind4 + 1][0]) / 2)

        elif ind4 == len(final_grouped_peaks) - 1:
            lower_point = int((group[0] + final_grouped_peaks[ind4 - 1][-1]) / 2)
            higher_point = len(total_spectral_ydata)

        else:
            lower_point = int((group[0] + final_grouped_peaks[ind4 - 1][-1]) / 2)

            higher_point = int((group[-1] + final_grouped_peaks[ind4 + 1][0]) / 2)

        final_peak_regions[ind4] = np.arange(lower_point, higher_point)

    # now simulate new regions and store region data

    final_sim_y = np.empty(len(final_grouped_peaks), dtype=object)
    for ind3, region in enumerate(final_peak_regions):
        fit_y = p7sim(total_params, region, final_grouped_peaks[ind3], ind3)

        final_sim_y[ind3] = fit_y
    # obligatory numpyfication
    final_peaks = np.array(final_peaks)
    final_grouped_peaks = np.array(
        [np.array(i) for i in final_grouped_peaks], dtype=object
    )
    # Close and join the pool to release resources
    pool.close()
    pool.join()

    return (
        final_peaks,
        final_grouped_peaks,
        final_peak_regions,
        final_sim_y,
        total_params,
    )



import logging
import time

import numpy as np
import nmrglue as ng

from dp5.nmr_processing.carbon import *
from .fid_corrections import spectral_processing
from .peak_picking import edge_removal, iterative_peak_picking
from .solvent_removal import solvent_removal
from .assign import iterative_assignment

from ..helper_functions import normalise_intensities


def carbon_processing(total_spectral_ydata, uc, solvent):
    total_spectral_ydata, spectral_ydata, threshold, corr_distance = (
        spectral_processing(total_spectral_ydata, uc)
    )

    spectral_xdata_ppm = uc.ppm_scale()

    total_spectral_ydata = edge_removal(total_spectral_ydata)

    picked_peaks, simulated_ydata = iterative_peak_picking(
        total_spectral_ydata, 5, corr_distance
    )

    picked_peaks = sorted(list(set(picked_peaks)))

    picked_peaks, removed = solvent_removal(
        simulated_ydata, spectral_xdata_ppm, solvent, uc, picked_peaks
    )

    return (
        spectral_xdata_ppm,
        total_spectral_ydata,
        picked_peaks,
        simulated_ydata,
        removed,
    )


def carbon_assignment(nmr_data, molecule, shifts, labels):
    """
    Arguments:
        nmr_data (:py:class:`~dp5.nmr_processing.nmr_ai.NMRData`:): NMR Data class
        molecule (:py:object:`rdkit.Chem.Mol`): rdkit molecule used to establish connectivity. Currently unused.
        shifts (:py:object:`numpy.ndarray`): numpy array of shifts
        label (:py:object:`numpy.ndarray`): numpy array of shifts
    """
    assigned_shifts, assigned_peaks, assigned_labels, scaled_shifts = (
        iterative_assignment(
            picked_peaks=nmr_data["exppeaks"],
            spectral_xdata_ppm=nmr_data["xdata"],
            total_spectral_ydata=nmr_data["ydata"],
            calculated_shifts=shifts,
            C_labels=labels,
        )
    )
    C_exp = [None] * len(shifts)
    for label, peak in zip(assigned_labels, assigned_peaks):

        w = labels.tolist().index(label)

        C_exp[w] = peak

    return np.array(C_exp)



import numpy as np
from numpy import linalg

"""
Performs conformer pruning. Returns a new list of conformers and RMSD cutoff
"""


def StrictRMSDPrune(conformers, cutoff: float, conf_limit: int):
    """
    Removes redundant conformers to fall below the specified conformer limit. Assumes conformers are sorted in ascending order by energy.

    Arguments:
    - conformers: a list of lists of lists of atomic coordinates. Follows the atom numbering from Isomer.InputFile.
    - atoms: a list of atomic symbols for atoms in Isomer. Follows the atom numbering from Isomer.InputFile. Redundant.
    - cutoff: a float. Conformers similar to lower energy conformers (RMSD<cutoff) are marked for deletion.
    - conf_limit: an integer. Number of conformers should not exceed this limit.

    Returns:
    - to_keep: a list of boolean values, where True corresponds to unique conformers.
    - cutoff: a float. The final value of the cutoff parameter.

    """

    def confs_to_keep(rmsd_matrix: np.array, cutoff: float):
        """
        Auxiliary function. Marks the non-redundant conformers. Assumes conformers are sorted in ascending order by energy.

        Arguments:
        - rmsd_matrix: an np.array containing aligned RMSD between conformers.
        - cutoff: a float. Conformers similar to lower energy conformers (RMSD<cutoff) are marked for deletion.
        Returns:
        - above_cutoff: a boolean array with indices set to True for retained conformers
        """
        # checks for RMSDs above the cutoff
        mask = rmsd_matrix >= cutoff

        """Sets the main diagonal and upper triangular elements as True. 
        Guarantees the retention of the redundant conformer encountered first.
        Higher energy redundant conformers should therefore have at least one False element in their column."""
        mask[np.triu_indices_from(mask)] = True
        above_cutoff = mask.all(axis=1)

        return above_cutoff

    conf_arrays = [np.array(conformer, dtype=np.float32) for conformer in conformers]
    num_confs = len(conformers)
    # align them all

    conf_zeroed = []
    for conf in conf_arrays:
        conf_zeroed.append(_Move2Origin(conf))

    rmsd_matrix = np.zeros((num_confs, num_confs))
    # populate the matrix
    for confid1 in range(0, num_confs):
        for confid2 in range(confid1 + 1, num_confs):
            res = _AlignedRMS(conf_zeroed[confid1], conf_zeroed[confid2])
            rmsd_matrix[confid1, confid2] = res
            rmsd_matrix[confid2, confid1] = res

    to_keep = confs_to_keep(rmsd_matrix, cutoff)

    while (to_keep.sum()) > conf_limit:
        cutoff += 0.2
        to_keep = confs_to_keep(rmsd_matrix, cutoff)

    return to_keep, cutoff


def _AlignedRMS(mol1, mol2):
    """
    aligns two conformers and estimates their RMSD. Does not take symmetry into account. Assumes conformers are translated to origin.

    arguments:
    - mol1: np.array of shape (3, number of atoms) containing coordinates for the first conformer
    - mol2: np.array of shape (3, number of atoms) containing coordinates for the second conformer

    returns:
        - root mean squared deviation
    """

    # align

    u = _qtrfit(mol1, mol2)
    mol1 = np.matmul(u, mol1.T).T

    # calculate RMS
    sd = (mol1 - mol2) ** 2

    return np.sqrt(sd.sum() / mol1.shape[0])


"""
 * This routine is a python translation of the QTRFIT Fortran routine
 * of David Heisterberg
 *
 * David J. Heisterberg
 * The Ohio Supercomputer Center, 1224 Kinnear Rd,
 * Columbus, OH  43212-1163
 * (614)292-6036; djh@osc.edu
 *
 * Python translation by K Ermanis 2015
 *
 * QTRFIT
 * Find the quaternion, q,[and left rotation matrix, u] 
 * that minimizes
 *   |qTBq - A| ^ 2   [|uB - A| ^ 2]
 * This is equivalent to maximizing Re (qTBTqA).
 * This is equivalent to finding the largest 
 * eigenvalue and corresponding
 * eigenvector of the matrix
 *  [V2   VUx  VUy  VUz ]
 *  [VUx  Ux2  UxUy UzUx]
 *  [VUy  UxUy Uy2  UyUz]
 *  [VUz  UzUx UyUz Uz2 ]
 * where
 *   V2   = Bx Ax + By Ay + Bz Az
 *   Ux2  = Bx Ax - By Ay - Bz Az
 *   Uy2  = By Ay - Bz Az - Bx Ax
 *   Uz2  = Bz Az - Bx Ax - By Ay
 *   VUx  = Bz Ay - By Az
 *   VUy  = Bx Az - Bz Ax
 *   VUz  = By Ax - Bx Ay
 *   UxUy = Bx Ay + By Ax
 *   UyUz = By Az + Bz Ay
 *   UzUx = Bz Ax + Bx Az
 * The left rotation matrix, u, is obtained from q by
 *   u = qT1q
 * INPUT
 *   n      - number of points
 *   b      - molecule to be rotated
 *   a      - reference molecule
 *   w      - weight vector
 * OUTPUT
 *   z      - eigenvalue
 *   q      - the best-fit quaternion
 *   u      - the best-fit left rotation matrix
 *   nr     - number of jacobi sweeps required
 
 KE:
 structure of molecule data:
 a[atomnr][coordinate]
 """


def _qtrfit(mol, refmol):
    """
    generates quaternion rotation matrix for mol such as its deviation from a reference molecule is minimised
    """
    # computes outer products of coordinates via a numpy trick, a being mol length, b, and c being coordinates
    product = np.einsum("ab,ac -> abc", mol, refmol)
    # outputs values
    xxyx, xxyy, xxyz, xyyx, xyyy, xyyz, xzyx, xzyy, xzyz = product.sum(axis=0).flatten()

    c = np.zeros((4, 4), dtype=np.float64)
    c[0, 0] = xxyx + xyyy + xzyz
    c[0, 1] = c[1, 0] = xzyy - xyyz
    c[1, 1] = xxyx - xyyy - xzyz
    c[0, 2] = c[2, 0] = xxyz - xzyx
    c[1, 2] = c[2, 1] = xxyy + xyyx
    c[2, 2] = xyyy - xzyz - xxyx
    c[0, 3] = c[3, 0] = xyyx - xxyy
    c[1, 3] = c[3, 1] = xzyx + xxyz
    c[2, 3] = c[3, 2] = xyyz + xzyy
    c[3, 3] = xzyz - xxyx - xyyy

    eval, evec = linalg.eigh(c)

    u = _q2mat(evec[3])

    return u


def _Move2Origin(mol: np.array):

    shift = mol.sum(axis=0) / mol.shape[0]
    mol = mol - shift

    return mol


def _q2mat(q: np.array):
    u = np.zeros((3, 3), dtype=np.float64)

    u[0, 0] = q[0] ** 2 + q[1] ** 2 - q[2] ** 2 - q[3] ** 2
    u[1, 0] = 2.0 * (q[1] * q[2] - q[0] * q[3])
    u[2, 0] = 2.0 * (q[1] * q[3] + q[0] * q[2])
    u[0, 1] = 2.0 * (q[2] * q[1] + q[0] * q[3])
    u[1, 1] = q[0] ** 2 - q[1] ** 2 + q[2] ** 2 - q[3] ** 2
    u[2, 1] = 2.0 * (q[2] * q[3] - q[0] * q[1])
    u[0, 2] = 2.0 * (q[3] * q[1] - q[0] * q[2])
    u[1, 2] = 2.0 * (q[3] * q[2] + q[0] * q[1])
    u[2, 2] = q[0] ** 2 - q[1] ** 2 - q[2] ** 2 + q[3] ** 2

    return u



""" Unused for the moment, I believe. """

import tensorflow as tf

def get_shape(tensor):
    """Returns the tensor's shape.
    Each shape element is either:
    - an `int`, when static shape values are available, or
    - a `tf.Tensor`, when the shape is dynamic.
    Args:
    tensor: A `tf.Tensor` to get the shape of.
    Returns:
    The `list` which contains the tensor's shape.
    """

    shape_list = tensor.shape.as_list()
    if all(s is not None for s in shape_list):
        return shape_list
    
    shape_tensor = tf.shape(tensor)
    return [shape_tensor[i] if s is None else s for i, s in
            enumerate(shape_list)]


def repeat(tensor, repeats, axis=0):
    """Repeats a `tf.Tensor`'s elements along an axis by custom amounts.
    Equivalent to Numpy's `np.repeat`.
    `tensor and `repeats` must have the same numbers of elements along `axis`.
    Args:
    tensor: A `tf.Tensor` to repeat.
    repeats: A 1D sequence of the number of repeats per element.
    axis: An axis to repeat along. Defaults to 0.
    name: (string, optional) A name for the operation.
    Returns:
    The `tf.Tensor` with repeated values.
    """

    cumsum = tf.cumsum(repeats)
    range_ = tf.range(cumsum[-1])

    indicator_matrix = tf.cast(tf.expand_dims(range_, 1) >= cumsum, tf.int32)
    indices = tf.reduce_sum(indicator_matrix, reduction_indices=1)

    shifted_tensor = _axis_to_inside(tensor, axis)
    repeated_shifted_tensor = tf.gather(shifted_tensor, indices)
    repeated_tensor = _inside_to_axis(repeated_shifted_tensor, axis)

    shape = tensor.shape.as_list()
    shape[axis] = None
    repeated_tensor.set_shape(shape)

    return repeated_tensor


def _axis_to_inside(tensor, axis):
    """Shifts a given axis of a tensor to be the innermost axis.
    Args:
        tensor: A `tf.Tensor` to shift.
        axis: An `int` or `tf.Tensor` that indicates which axis to shift.
    Returns:
        The shifted tensor.
    """

    axis = tf.convert_to_tensor(axis)
    rank = tf.rank(tensor)

    range0 = tf.range(0, limit=axis)
    range1 = tf.range(tf.add(axis, 1), limit=rank)
    perm = tf.concat([[axis], range0, range1], 0)

    return tf.transpose(tensor, perm=perm)


def _inside_to_axis(tensor, axis):
    """Shifts the innermost axis of a tensor to some other axis.
    Args:
        tensor: A `tf.Tensor` to shift.
        axis: An `int` or `tf.Tensor` that indicates which axis to shift.
    Returns:
        The shifted tensor.
    """

    axis = tf.convert_to_tensor(axis)
    rank = tf.rank(tensor)

    range0 = tf.range(1, limit=axis + 1)
    range1 = tf.range(tf.add(axis, 1), limit=rank)
    perm = tf.concat([range0, [0], range1], 0)

    return tf.transpose(tensor, perm=perm)



""" Keras loss functions will choke on NaN inputs, which we'll often have for
unspecified inputs. These are just a couple simple loss functions that mask NaN
values in both the test and predicted tensors when computing the cost. """

import keras.backend as K
import tensorflow as tf

def masked_mean_squared_error(y_true, y_pred):
    mask = tf.is_finite(y_true)
    y_true_mask = tf.boolean_mask(y_true, mask)
    y_pred_mask = tf.boolean_mask(y_pred, mask)
    return K.mean(K.square(y_pred_mask - y_true_mask), axis=-1)


def masked_mean_absolute_error(y_true, y_pred):
    mask = tf.is_finite(y_true)
    y_true_mask = tf.boolean_mask(y_true, mask)
    y_pred_mask = tf.boolean_mask(y_pred, mask)
    return K.mean(K.abs(y_pred_mask - y_true_mask), axis=-1)



import numpy as np
from rdkit import Chem
from rdkit.Geometry import Point3D
from rdkit.Chem import rdForceFieldHelpers

from dp5.conformer_search.run_cs import conf_search
from dp5.dft.run_dft import dft_calculations
from dp5.neural_net.nn_utils import get_nn_shifts
from dp5.analysis.dp5 import DP5
from dp5.analysis.dp4 import DP4


class Molecule:
    def __init__(self, input_file: str):
        self.input_file = input_file
        self.base_name = input_file.rsplit(".", maxsplit=1)[0]
        mol = Chem.MolFromMolFile(input_file, removeHs=False)

        self.atoms = [at.GetSymbol() for at in mol.GetAtoms()]
        self.conformers = [mol.GetConformer(0).GetPositions()]
        self.charge = sum([at.GetFormalCharge() for at in mol.GetAtoms()])

        # estimates force field energy
        prop = rdForceFieldHelpers.MMFFGetMoleculeProperties(mol, mmffVariant="MMFF94s")
        ff = rdForceFieldHelpers.MMFFGetMoleculeForceField(mol, prop)
        self._energies = np.array([float(ff.CalcEnergy()) * 4.184])
        # creates mol object for further manipulation
        self._rdkit_mols = None
        self._populations = None
        self._mol = mol

    def __repr__(self) -> str:
        return self.base_name

    @property
    def atoms(self):
        return self._atoms

    @atoms.setter
    def atoms(self, new_atoms):
        self._atoms = np.array(new_atoms)

    @property
    def conformers(self):
        return self._conformers

    @conformers.setter
    def conformers(self, values):
        if not isinstance(values, np.ndarray):
            values = np.array(values)
        if values.shape[2] == 3:
            self._conformers = values
        else:
            raise ValueError("Cannot set coordinates!")
        self._rdkit_mols = None

    @property
    def energies(self):
        return self._energies

    @energies.setter
    def energies(self, values):
        if not isinstance(values, np.ndarray):
            values = np.array(values)
        self._energies = values
        self._populations = None

    @property
    def charge(self):
        return self._charge

    @charge.setter
    def charge(self, value):
        self._charge = int(value)

    @property
    def rdkit_mols(self):
        if self._rdkit_mols is None:
            mols = []
            for conformer in self.conformers:
                molecule = Chem.Mol(self._mol)
                conf = molecule.GetConformer(0)
                for atom, atom_coord in enumerate(conformer):
                    x, y, z = atom_coord
                    conf.SetAtomPosition(atom, Point3D(x, y, z))
                mol = Chem.Mol(molecule, confId=0)
                mols.append(mol)
            self._rdkit_mols = mols
        return self._rdkit_mols

    @property
    def conformer_C_pred(self):
        return self._conformer_C_pred

    @conformer_C_pred.setter
    def conformer_C_pred(self, values):
        if not isinstance(values, np.ndarray):
            values = np.array(values)
        self._conformer_C_pred = values

    @property
    def C_labels(self):
        return self._C_labels

    @C_labels.setter
    def C_labels(self, labels):
        self._C_labels = np.array(labels)

    @property
    def H_labels(self):
        return self._H_labels

    @H_labels.setter
    def H_labels(self, labels):
        self._H_labels = np.array(labels)

    @property
    def shielding_labels(self):
        return self._shielding_labels

    @shielding_labels.setter
    def shielding_labels(self, labels):
        self._shielding_labels = np.array(labels)

    def add_conformer_data(self, data):
        self.atoms = data.atoms
        self.conformers = data.conformers
        self.charge = data.charge
        self.energies = data.energies
        self._rdkit_mols = None

    def add_dft_data(self, data):
        attrs = (
            "atoms",
            "conformers",
            "charge",
            "energies",
            "conformer_C_pred",
            "C_labels",
            "conformer_H_pred",
            "H_labels",
            "shielding_labels",
        )
        for attr in attrs:
            if hasattr(data, attr):
                setattr(self, attr, getattr(data, attr))
        self._rdkit_mols = None

    def add_nn_shifts(self, shifts_labels):
        self.conformer_C_pred, self.C_labels, self.conformer_H_pred, self.H_labels = (
            shifts_labels
        )

    def copy(self):
        """Prevents accidental attribute override."""
        new_mol = type(self).__new__(self.__class__)
        new_mol.__dict__.update(self.__dict__)

        return new_mol

    @property
    def populations(self):
        """Assumeds kJ/mol as energy unit"""
        if self._populations is None:
            energies = self.energies
            scaling = 1000 / 8.3415 / 298.15
            energies = energies - np.min(energies)
            exp_energies = np.exp(-energies * scaling)
            self._populations = exp_energies / exp_energies.sum()
        return self._populations

    def boltzmann_weighting(self, attr: str):
        # recomputes populations just in case
        data = getattr(self, attr)
        data = np.array(data, dtype=np.float32)
        return (self.populations[:, np.newaxis] * data).sum(axis=0)

    @property
    def H_shifts(self):
        return self.boltzmann_weighting("conformer_H_pred")

    @property
    def C_shifts(self):
        return self.boltzmann_weighting("conformer_C_pred")

    def assign_nmr(self, C_exp, H_exp):
        self.C_exp, self.H_exp = np.array(C_exp, dtype=np.float32), np.array(
            H_exp, dtype=np.float32
        )

    def add_dp4_data(self, dp4_data):
        self.dp4_data = dp4_data

    def add_dp5_data(self, dp5_data):
        self.dp5_data = dp5_data


class Molecules:
    """Class that handles all the calculations. Should keep the molecular data in itself"""

    def __init__(self, config):
        self.config = config
        self.mols = [Molecule(mol) for mol in self.config["structure"]]

    def __iter__(self):
        return (mol.copy() for mol in self.mols)

    def __getitem__(self, idx):
        return self.mols[idx]

    def get_conformers(self):
        """Runs conformational search."""
        mm_data = conf_search(self.mols, self.config["conformer_search"])
        for mol, data in zip(self.mols, mm_data):
            mol.add_conformer_data(data)

    def get_dft_data(self):
        """Runs DFT calculations"""
        dft_mols = [mol for mol in self.mols]
        dft_data = dft_calculations(
            dft_mols, self.config["workflow"], self.config["dft"]
        )
        for mol, data in zip(self.mols, dft_data):
            mol.add_dft_data(data)

    def get_nn_nmr_shifts(self):
        """Should get C and H shifts"""
        mols = [mol.rdkit_mols for mol in self.mols]
        cascade_shifts_labels = get_nn_shifts(mols)
        for mol, *m_shift_label in zip(self.mols, *cascade_shifts_labels):
            mol.add_nn_shifts(m_shift_label)

    def assign_nmr_spectra(self, nmrdata):
        for mol in self.mols:
            C_exp, H_exp = nmrdata.assign(mol)
            mol.assign_nmr(C_exp, H_exp)

    def dp5_analysis(self):
        dp5 = DP5(self.config["output_folder"], self.config["workflow"]["dft_nmr"])
        self.dp5_output = dp5(self.mols)

    def dp4_analysis(self):
        dp4 = DP4(self.config["output_folder"], self.config["dp4"])
        self.dp4_output = dp4(self.mols)

    def print_results(self):
        output = "Workflow summary\n\n"
        output += f"Solvent = {self.config['solvent']}\n"
        if self.config["workflow"]["dft_opt"]:
            output += (
                f"DFT optimisation functional: {self.config['dft']['o_functional']}\n"
                f"DFT optimisation basis set: {self.config['dft']['o_basis_set']}\n"
            )
        if self.config["workflow"]["dft_energies"]:
            output += (
                f"DFT energy functional: {self.config['dft']['e_functional']}\n"
                f"DFT energy basis set: {self.config['dft']['e_basis_set']}\n"
            )
        if self.config["workflow"]["dft_nmr"]:
            output += (
                f"DFT NMR functional: {self.config['dft']['n_functional']}\n"
                f"DFT NMR basis set: {self.config['dft']['n_basis_set']}\n"
            )
        if self.config["dp4"]["param_file"] != "none":
            output += (
                f"\nDP4 statistical model file: {self.config['dp4']['param_file']}\n"
            )

        output += f"\nNumber of candidates: {len(self.mols)}\n"

        for i, mol in enumerate(self.mols):
            output += (
                f"Number of conformers for molecule {mol}: {len(mol.conformers)}\n"
            )

        if self.config["workflow"]["dp4"]:
            dp4_output = output + self.dp4_output
            with open((self.config["output_folder"]) / "output.dp4", "w") as f:
                f.write(dp4_output)

        if self.config["workflow"]["dp5"]:
            dp5_output = output + self.dp5_output
            with open((self.config["output_folder"]) / "output.dp5", "w") as f:
                f.write(dp5_output)



import logging

import numpy as np
from scipy.stats import norm, gmean

from dp5.nmr_processing.helper_functions import proton_count, methyl_protons, labile_protons

logger = logging.getLogger(__name__)


def find_integrals(
    mol,
    peak_regions,
    grouped_peaks,
    sim_regions,
    picked_peaks,
    params,
    y_data,
):
    """
    - mol: rdkit Mol object
    - peak_regions: the regions where the peaks are found
    - grouped_peaks: peak grouping for subsequent integration
    - sim_regions: regions to simulate
    - picked_peaks: identified peaks
    - params: params
    - y_data: normalised intensity
    """

    # count number of protons in the file
    structure_protons = proton_count(mol)

    # find number of methyl groups
    m = methyl_protons(mol)
    number_of_methyl_groups = len(m)

    # count the number of labile protons in the structure
    l_protons = labile_protons(mol)

    logger.info(f"number of protons = {structure_protons}")
    logger.info(f"number of labile protons = {l_protons}")

    count = 0

    # allow guesses of number of protons in the spectrum between: structure_protons - l_protons, 2 * structure_protons

    number_vector = np.arange(
        structure_protons - l_protons, 2 * structure_protons)

    scores = np.zeros(len(number_vector))

    integrals = integrate_sim_regions(
        sim_regions, grouped_peaks, peak_regions, y_data, params
    )

    for proton_guess in number_vector:

        norm_integrals = normalise_integration(integrals, proton_guess)

        # remove impurities

        # find number of impurities removed

        # find which should be classed as impurities

        r = sum_round(norm_integrals)

        r = np.array(r)

        impurities = len(r[r < 1])

        norm_integrals = norm_integrals[r > 0.5]

        scores[count] = integral_score(
            norm_integrals, structure_protons, proton_guess, l_protons, impurities
        )

        number_of_methyl_groups_integral = np.sum((r - (r % 3)) // 3)

        if number_of_methyl_groups_integral < number_of_methyl_groups:
            scores[count] = 0

        count += 1

    wh = np.argmax(scores)

    best_fit = number_vector[wh]

    logger.info("the best fit number of protons is " + str(best_fit))

    # normalise using this number

    integrals = normalise_integration(integrals, best_fit)

    (
        grouped_peaks,
        integrals,
        peak_regions,
        picked_peaks_,
        impurities,
        sim_regions,
        rounded_integrals,
    ) = remove_impurities(
        integrals, peak_regions, grouped_peaks, picked_peaks, sim_regions
    )

    integral_sum, cummulative_vectors = integral_add(sim_regions, best_fit)

    total_r = 0

    total = 0

    for i in range(0, len(integrals)):
        total += integrals[i]
        total_r += rounded_integrals[i]

    return (
        peak_regions,
        grouped_peaks,
        sim_regions,
        integral_sum,
        cummulative_vectors,
        rounded_integrals,
        structure_protons,
        best_fit,
        total_r,
    )


def integral_score(integrals, structure_protons, proton_guess, l_protons, impurities):

    r_int = sum_round(integrals)

    r_int = np.array(r_int)

    sum_r = np.sum(r_int)

    new_integrals = []

    for ind, integral in enumerate(integrals):
        new_integrals += [integral] * int(r_int[ind])

    new_integrals = np.array(new_integrals)

    differences = new_integrals % 0.5

    std = structure_protons / 8

    diff = abs(proton_guess - structure_protons)

    probs = 4 * (1 - norm.cdf(differences, loc=0, scale=1 / 16))

    mean = gmean(probs) * (1 - norm.cdf(diff, loc=0, scale=std)) * \
        (1 / 2**impurities)

    # only allow intergrals that are between the expected number and that number - the number of labile protons

    if sum_r < structure_protons - l_protons:
        mean = 0

    return mean


def integral_add(sim_regions, proton_guess):

    integrals = [np.sum(sim) for sim in sim_regions]

    i_sum = np.sum(integrals)

    Inorm = proton_guess / i_sum

    integrals_copy = list(integrals)

    integrals_copy = [0] + integrals_copy

    integral_sum = np.cumsum(np.asarray(integrals_copy))
    integral_sum = integral_sum / np.sum(integrals_copy)

    cummulative_vectors = []

    t_sum = 0
    for region in sim_regions:
        t_sum += np.sum(region)
        cummulative_vectors.append(np.cumsum(region))

    for i, region in enumerate(cummulative_vectors):
        cummulative_vectors[i] = region / t_sum

    return integral_sum, cummulative_vectors


def normalise_integration(integrals, initial_proton_guess):
    i_sum = np.sum(integrals)
    integrals = integrals / i_sum
    norm_integrals = integrals * initial_proton_guess

    return norm_integrals


def integrate_sim_regions(
    sim_regions, grouped_peaks, peak_regions, y_data, params
):
    sim_integrals = []

    for r, group in enumerate(grouped_peaks):

        region_integral = 0

        for peak in group:
            region_integral += (
                params["A" + str(peak)]
                * 0.25
                * np.pi
                * params["std" + str(peak)]
                * ((3**0.5 - 2) * params["vregion" + str(r)] + 2)
            )

        sim_integrals.append(region_integral)

    sim_integrals = np.array(sim_integrals)
    y_integral = np.sum(y_data)
    sim_integral = np.sum(sim_integrals)
    k = sim_integral / y_integral
    integrals = []

    for region in peak_regions:
        integrals.append(np.sum(y_data[region]))

    k = np.array(sim_integrals) / np.array(integrals)

    simr_regions = []

    for region in sim_regions:
        simr_regions.append(np.sum(region))

    integrals = k * integrals

    return integrals


def remove_impurities(
    integrals, peak_regions, grouped_peaks, picked_peaks, sim_regions
):
    '''
    - integrals
    - peak_regions
    - grouped_peaks
    - picked_peaks
    - sim_regions
    currently suffers from ragged arrays and numpy not nadling them: must be fixed upstream
    '''
    # find rounded values

    r = sum_round(integrals)

    r = np.array(r)

    to_remove = np.where(r < 0.5)[0]

    number_of_impurities = len(to_remove)

    peaks_to_remove = []
    for group in to_remove:
        peaks_to_remove.extend(grouped_peaks[group])

    whdel = np.isin(picked_peaks, peaks_to_remove)

    picked_peaks = np.delete(picked_peaks, whdel)

    integrals = np.delete(integrals, to_remove)

    peak_regions = np.delete(peak_regions, to_remove)

    grouped_peaks = np.delete(grouped_peaks, to_remove)

    sim_regions = np.delete(sim_regions, to_remove)

    rounded_integrals = r[r > 0.5]

    return (
        grouped_peaks,
        integrals,
        peak_regions,
        picked_peaks,
        number_of_impurities,
        sim_regions,
        rounded_integrals,
    )


def sum_round(a):
    """
    Round each element in the input list to the nearest integer and adjust some
    values to minimize the error in the sum of the rounded values. Used for proton integration

    Parameters:
    - a (list of float): Input list of numeric values.

    Returns:
    - list of int: A list of integers where each element is the result of rounding
      the corresponding element in the input list to the nearest integer, with some
      adjustments to minimize the error in the sum.
    """

    # Round each element in the input list 'a' to the nearest integer
    rounded = [round(i, 0) for i in a]

    # Calculate the sum of the original list 'a' and the sum of the rounded values
    error = sum(a) - sum(rounded)

    # Calculate the integer part of the 'error' divided by 1
    n = int(round(error / 1))

    # Create a copy of the rounded list
    new = rounded[:]

    # Sort the list of tuples containing the difference between elements in 'a' and 'rounded',
    # along with the index of the element, in descending order.
    # Select the top 'abs(n)' elements from the sorted list.
    for _, i in sorted(((a[i] - rounded[i], i) for i in range(len(a))), reverse=n > 0)[
        : abs(n)
    ]:
        # Adjust the corresponding element in the 'new' list by adding 'n / abs(n)'
        new[i] += n / abs(n)

    # Return the modified list 'new'
    return new


def weighted_region_centres(peak_regions, total_spectral_ydata):
    centres = []

    for region in peak_regions:
        w = total_spectral_ydata[region] ** 2

        wx = region * w

        xbar = np.sum(wx) / np.sum(total_spectral_ydata[region] ** 2)

        centres.append(int(xbar))

    return centres
